"""
Command-line utility for opening the frequency support analysis tools.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

from sarpy_apps.apps.frequency_support_tool.full_support_tool import main as full_main
from sarpy_apps.apps.frequency_support_tool.local_support_tool import main as local_main


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser(
        description="Open one of the frequency support tools with optional input file.",
        formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument(
        '-i', '--input', metavar='input', default=None, help='The path to the optional image file for opening.')
    parser.add_argument(
        '-t', '--type', default='local', choices=['local', 'full'],
        help="What type of analysis is being performed?")
    args = parser.parse_args()

    if args.type == 'local':
        local_main(reader=args.input)
    elif args.type == 'full':
        full_main(reader=args.input)
    else:
        raise ValueError('Got unexpected type {}'.format(args.type))



"""
CRSD pulse explorer
"""

__classification__ = "UNCLASSIFIED"
__author__ = ("Thomas Rackers", "Thomas McCullough")

import logging
import os

import numpy
from scipy.signal import spectrogram, resample

import tkinter
from tkinter import ttk
from tkinter.messagebox import showinfo

from tkinter.filedialog import askopenfilename

from tk_builder.base_elements import TypedDescriptor, StringDescriptor, \
    BooleanDescriptor, IntegerDescriptor, FloatDescriptor
from tk_builder.panels.pyplot_image_panel import PyplotImagePanel
from tk_builder.widgets.basic_widgets import Frame, Label, Entry, Combobox, Button, Scale

from sarpy_apps.supporting_classes.file_filters import crsd_files, all_files
from sarpy_apps.supporting_classes.widget_with_metadata import \
    WidgetWithMetadata

from sarpy.io.general.base import SarpyIOError, FlatReader
from sarpy.io.received.converter import open_received
from sarpy.io.received.base import CRSDTypeReader
from sarpy.processing.windows import kaiser
from sarpy.processing.fft_base import fftshift

from sarpy_apps.supporting_classes.image_reader import CRSDTypeCanvasImageReader

logger = logging.getLogger(__name__)

_PULSE_DISPLAY_VALUES = ('RFSignal',)


def _reramp(pulse_data, sampling_rate, deramp_rate):
    """

    Parameters
    ----------
    pulse_data : numpy.ndarray
    sampling_rate : float
    deramp_rate : float

    Returns
    -------
    (numpy.ndarray, float)
        The reramped_data and new sampling rate.
    """

    rcv_window_length = pulse_data.size / sampling_rate
    deramp_bandwidth = abs(rcv_window_length * deramp_rate)
    upsample_factor = (deramp_bandwidth + sampling_rate) / sampling_rate

    sample_size = round(upsample_factor * pulse_data.size)
    oversample_factor = sample_size / (upsample_factor * pulse_data.size)
    use_data = resample(pulse_data - numpy.mean(pulse_data), sample_size)
    time_interval = 1 / (sampling_rate * upsample_factor * oversample_factor)
    times = time_interval * numpy.arange(sample_size) - 0.5 * rcv_window_length
    return use_data * numpy.exp(
        1j * numpy.pi * deramp_rate * times * times), 1. / time_interval


def _stft(data, sampling_rate):
    """
    Take the specified short time fourier transform of the provided data.

    Parameters
    ----------
    data : numpy.ndarray
        Must be one-dimensional
    sampling_rate : float
        The sampling rate.

    Returns
    -------
    numpy.ndarray
    """

    if not (isinstance(data, numpy.ndarray) and data.ndim == 1):
        raise ValueError(
            'This short time fourier transform function only applies to a '
            'one-dimensional numpy array')

    nfft = 2 ** int(numpy.ceil(numpy.log2(numpy.sqrt(data.shape[0]))) + 2)
    nperseg = int(0.97 * nfft)
    window = kaiser(nperseg, 5)
    noverlap = int(0.9 * nfft)
    frequencies, times, trans_data = spectrogram(
        data, sampling_rate, window=window, nperseg=nperseg, noverlap=noverlap,
        nfft=nfft, return_onesided=False)
    return times, fftshift(frequencies, axes=0), fftshift(trans_data, axes=0)


def _rf_signal(reader, index, pulse):
    """
    Gets the RF Signal data for the given index (channel) and pulse.

    Parameters
    ----------
    reader : CRSDTypeReader
    index : int
    pulse : int

    Returns
    -------
    (numpy.ndarray, numpy.ndarray, numpy.ndarray)
        The times, frequencies, and stft data array
    """

    crsd = reader.crsd_meta
    params = crsd.Channel.Parameters[index]
    sampling_rate = params.Fs
    pulse_data = reader[pulse, :, index].flatten()
    fic_rate = float(reader.read_pvp_variable('FICRate', index, pulse)[0])
    dfic0 = float(reader.read_pvp_variable('DFIC0', index, pulse)[0])

    if fic_rate == 0:
        times, frequencies, stft_data = _stft(pulse_data, sampling_rate)
        frequencies += params.F0Ref + dfic0 - 0.5 * sampling_rate
    else:
        reramped, reramped_sampling_rate = _reramp(pulse_data, sampling_rate,
                                                   fic_rate)
        times, frequencies, stft_data = _stft(reramped, reramped_sampling_rate)
        frequencies += \
            params.F0Ref + dfic0 + \
            0.5 * fic_rate * pulse_data.size / reramped_sampling_rate - \
            0.5 * reramped_sampling_rate
    return times, frequencies, stft_data


class STFTCanvasImageReader(CRSDTypeCanvasImageReader):
    __slots__ = (
        '_base_reader', '_chippers', '_index', '_data_size', '_remap_function',
        '_signal_data_size', '_pulse', '_pulse_display', '_pulse_data',
        '_times', '_frequencies')

    def __init__(self, reader):
        """

        Parameters
        ----------
        reader : str|CRSDTypeReader
            The crsd type reader, or path to appropriate data file.
        """

        self._signal_data_size = None
        self._pulse = None
        self._pulse_display = None
        self._pulse_data = None
        self._times = None
        self._frequencies = None
        self.pulse_display = _PULSE_DISPLAY_VALUES[0]
        CRSDTypeCanvasImageReader.__init__(self, reader)

    @property
    def pulse_display(self):
        """
        str: The pulse display string determining methodology
        """

        return self._pulse_display

    @pulse_display.setter
    def pulse_display(self, value):
        if value not in _PULSE_DISPLAY_VALUES:
            raise ValueError(
                'pulse display (`{}`) not one of the allowed values\n\t`{}`'.format(
                    value, _PULSE_DISPLAY_VALUES))
        self._pulse_display = value
        self._set_pulse_data()

    @property
    def pulse_count(self):
        """
        None|int: The pulse count.
        """

        if self._signal_data_size is None:
            return None
        return self._signal_data_size[0]

    @property
    def index(self):
        """
        int: The reader index.
        """

        return self._index

    @index.setter
    def index(self, value):
        value = int(value)
        signal_data_sizes = self.base_reader.get_data_size_as_tuple()
        if not (0 <= value < len(signal_data_sizes)):
            logging.error(
                'The index property must be 0 <= index < {}, '
                'and got argument {}. Setting to 0.'.format(
                    len(signal_data_sizes), value))
            value = 0
        self._index = value
        self._signal_data_size = signal_data_sizes[value]
        self.pulse = 0

    @property
    def base_reader(self):
        # type: () -> CRSDTypeReader
        """
        CRSDTypeReader: The crsd reader object
        """

        return self._base_reader

    @base_reader.setter
    def base_reader(self, value):
        if isinstance(value, str):
            reader = None
            try:
                reader = open_received(value)
            except SarpyIOError:
                pass

            if reader is None:
                raise SarpyIOError(
                    'Could not open file {} as a CRSD reader'.format(value))
            value = reader

        if not isinstance(value, CRSDTypeReader):
            raise TypeError(
                'base_reader must be a CRSDReader, got type {}'.format(
                    type(value)))
        self._base_reader = value
        # noinspection PyProtectedMember
        self._chippers = value._get_chippers_as_tuple()
        self.index = 0

    @property
    def pulse(self):
        """
        int: The pulse currently being considered.
        """

        return self._pulse

    @pulse.setter
    def pulse(self, value):
        if self.pulse_count is None:
            self._pulse = None
            self._pulse_data = None

        value = int(value)
        # value should be between 0 and (pulse_count - 1).
        if not (0 <= value < self.pulse_count):
            raise ValueError('Pulse must be in range [0, {})'.format(self.pulse_count))

        self._pulse = value
        self._set_pulse_data()

    @property
    def times(self):
        """
        numpy.ndarray: Gets the times array for stft/spectrogram data for the current pulse.
        """

        return self._times

    @property
    def frequencies(self):
        """
        numpy.ndarray: Gets the frequencies array for stft/spectrogram data for the current pulse.
        """

        return self._frequencies

    @property
    def pulse_data(self):
        """
        numpy.ndarray: The spectrogram of the currently selected pulse
        """
        return self._pulse_data

    def _set_pulse_data(self):
        if self.pulse is None:
            self._pulse_data = None
            return

        if self._pulse_display == 'RFSignal':
            times, frequencies, data = _rf_signal(self.base_reader, self.index, self.pulse)
        else:
            raise ValueError(
                'Got unhandled pulse display value `{}`'.format(self.pulse_display))

        self._times = times
        self._frequencies = frequencies
        self._data_size = data.shape
        self._pulse_data = FlatReader(data)

    def __getitem__(self, item):
        if self._pulse_data is None:
            return None
        if self.remap_function is None:
            return self._pulse_data.__getitem__(item)
        return self.remap_complex_data(self._pulse_data.__getitem__(item))


###########
# The main app

class AppVariables(object):
    """
    App variables for the aperture tool.
    """

    browse_directory = StringDescriptor(
        'browse_directory', default_value=os.path.expanduser('~'),
        docstring='The directory for browsing for file selection.')  # type: str
    image_reader = TypedDescriptor(
        'image_reader', STFTCanvasImageReader,
        docstring='The crsd type canvas image reader object.')  # type: STFTCanvasImageReader
    animating = BooleanDescriptor(
        'animating', default_value=False,
        docstring='Are we currently looping through pulses?')  # type: bool
    animation_delay = IntegerDescriptor(
        'animation_delay', default_value=30, docstring='Animation delay in milliseconds')  # type: int
    vmin = FloatDescriptor('vmin', default_value=0)  # type: float
    vmax = FloatDescriptor('vmax', default_value=0)  # type: float
    vcount = IntegerDescriptor('vcount', default_value=0)  # type: int


class SliderWidget(Frame):
    """
    Widget panel with slider for forward/reverse and manual scan.
    """

    def __init__(self, parent):
        """

        Parameters
        ----------
        parent: Frame
        """

        Frame.__init__(self, parent)

        self['padding'] = 5

        self.start_pulse_label = Label(self, text='0')

        self.label_channel = Label(self, text='Channel')
        self.cbx_channel = Combobox(self, state='readonly')
        self.var_cbx_channel = tkinter.StringVar()
        self.cbx_channel.configure(
            justify='left', textvariable=self.var_cbx_channel,
            width=30, values=[], state='disabled')

        self.label_pulse = Label(self, text='Pulse')
        self.entry_pulse = Entry(self)
        self.entry_pulse.set_text('0')
        self.entry_pulse.configure(justify='right', width=6)
        self.entry_callback = self.register(self._only_numeric_input)
        self.entry_pulse.configure(validate="key", validatecommand=(self.entry_callback, "%P"))

        self.fullscale = Label(self, text='0')

        self.start_pulse_label.grid(row=0, column=0, padx=5, sticky='w')
        self.label_channel.grid(row=0, column=1, padx=5, sticky='e')
        self.cbx_channel.grid(row=0, column=2, padx=5)
        self.label_pulse.grid(row=0, column=3, padx=5)
        self.entry_pulse.grid(row=0, column=4, padx=5, sticky='w')
        self.fullscale.grid(row=0, column=5, padx=5, sticky='e')
        self.columnconfigure(0, weight=0)
        self.columnconfigure(1, weight=1)
        self.columnconfigure(2, weight=0)
        self.columnconfigure(3, weight=0)
        self.columnconfigure(4, weight=1)
        self.columnconfigure(5, weight=0)

        self.var_pulse_number = tkinter.StringVar(value='0')
        self.scale_pulse = Scale(
            self, from_=0, to=0, length=600, orient='horizontal',
            command=lambda s: self.var_pulse_number.set(s))
        self.scale_pulse.configure(variable=self.var_pulse_number)
        self.scale_pulse.grid(
            row=1, column=0, columnspan=6, padx=5, pady=8, sticky='esw')

    @staticmethod
    def _only_numeric_input(p):
        if p.isdigit() or p == "":
            return True
        return False


class DirectionWidget(Frame):
    """
    Button pair to select forward/reverse/stopped direction.
    """

    def __init__(self, parent):
        """

        Parameters
        ----------
        parent
        """

        Frame.__init__(self, parent)

        self.parent = parent

        self.button_style = ttk.Style()
        self.button_style.configure(
            'ToggleOff.TButton', font=('Arial', 24), foreground="gray50",
            background="PaleTurquoise3", width=3, sticky='CENTER')
        self.button_style.configure(
            'ToggleOn.TButton', font=('Arial', 24), foreground="black",
            background="PaleTurquoise1", width=3, sticky='CENTER')

        self.button_rev = Button(self.parent, text="\u25C0", takefocus=0, style='ToggleOff.TButton')
        self.button_prev = Button(self.parent, text="-1", takefocus=0, style='ToggleOff.TButton')
        self.button_next = Button(self.parent, text="+1", takefocus=0, style='ToggleOff.TButton')
        self.button_fwd = Button(self.parent, text="\u25B6", takefocus=0, style='ToggleOff.TButton')


class PulseExplorer(Frame, WidgetWithMetadata):
    def __init__(self, primary, reader=None, **kwargs):
        """

        Parameters
        ----------
        primary : tkinter.Toplevel|tkinter.Tk
        reader : None|str|CRSDTypeReader|CRSDTypeCanvasImageReader
        kwargs
            Keyword arguments passed through to the Frame
        """

        self.root = primary
        self.variables = AppVariables()

        Frame.__init__(self, primary, **kwargs)
        WidgetWithMetadata.__init__(self, primary)

        self.pyplot_panel = PyplotImagePanel(self, navigation=True)  # type: PyplotImagePanel

        self.pyplot_panel.cmap_name = 'turbo'
        self.pyplot_panel.set_ylabel('Freq (GHz)')
        self.pyplot_panel.set_xlabel('Time (\u03BCsec)')
        self.pyplot_panel.set_title('Pulse Visualization')

        self.scanner_panel = Frame(self, padding=10)  # type: Frame
        self.scanner_panel.columnconfigure(0, weight=0)
        self.scanner_panel.columnconfigure(1, weight=0)
        self.scanner_panel.columnconfigure(2, weight=1)
        self.scanner_panel.columnconfigure(3, weight=0)
        self.scanner_panel.columnconfigure(4, weight=0)

        self.dir_buttons = DirectionWidget(self.scanner_panel)  # type: DirectionWidget
        self.slider = SliderWidget(self.scanner_panel)  # type: SliderWidget

        self.dir_buttons.button_rev.grid(row=0, column=0, sticky='w')
        self.dir_buttons.button_prev.grid(row=0, column=1, sticky='w')
        self.slider.grid(row=0, column=2, sticky='ew')
        self.dir_buttons.button_next.grid(row=0, column=3, sticky='e')
        self.dir_buttons.button_fwd.grid(row=0, column=4, sticky='e')

        self.pyplot_panel.pack(side="top", expand=True, fill='both')
        self.scanner_panel.pack(side="bottom", expand=False, fill='x')

        self.rowconfigure(0, weight=1)
        self.rowconfigure(1, weight=0)
        self.pack(expand=True, fill='both')
        self.set_title()
        self._refresh_vdata()

        # define menus
        menubar = tkinter.Menu()
        # file menu
        filemenu = tkinter.Menu(menubar, tearoff=0)
        filemenu.add_command(label="Open Image", command=self.callback_select_files)
        filemenu.add_separator()
        filemenu.add_command(label="Exit", command=self.exit)
        # menus for informational popups
        popups_menu = tkinter.Menu(menubar, tearoff=0)
        popups_menu.add_command(label="Metaicon", command=self.metaicon_popup)
        popups_menu.add_command(label="Metaviewer", command=self.metaviewer_popup)
        # ensure menus cascade
        menubar.add_cascade(label="File", menu=filemenu)
        menubar.add_cascade(label="Metadata", menu=popups_menu)

        primary.config(menu=menubar)

        self.update_reader(reader)

        # self.image_panel.canvas.bind('<<RemapChanged>>', self.handle_remap_change)
        self.slider.cbx_channel.bind('<<ComboboxSelected>>', self.handle_image_index_changed)
        self.slider.scale_pulse.bind('<ButtonRelease>', self.handle_pulse_changed_slider)
        self.slider.entry_pulse.bind('<FocusOut>', self.handle_pulse_changed_entry)
        self.slider.entry_pulse.on_enter_or_return_key(self.handle_pulse_changed_entry)

        self.dir_buttons.button_rev.config(command=self.pulse_animate_backwards)
        self.dir_buttons.button_prev.config(command=self.pulse_step_prev)
        self.dir_buttons.button_next.config(command=self.pulse_step_next)
        self.dir_buttons.button_fwd.config(command=self.pulse_animate_forward)

    def set_title(self):
        """
        Sets the window title.
        """

        file_name = None if self.variables.image_reader is None \
            else self.variables.image_reader.file_name
        if file_name is None:
            the_title = "Pulse Explorer"
        else:
            the_title = "Pulse Explorer for {}".format(
                os.path.split(file_name)[1])
        self.winfo_toplevel().title(the_title)

    def set_pulse(self, value, override=False):
        """
        Sets the pulse count.

        Parameters
        ----------
        value : int
        override : bool
        """

        if self.variables.image_reader is None:
            return

        value = int(value)
        if value == self.variables.image_reader.pulse and not override:
            return

        if not (0 <= value < self.variables.image_reader.pulse_count):
            showinfo(
                'Invalid pulse',
                message='Pulse must be in the range [0, {})'.format(self.variables.image_reader.pulse_count))
            return

        self.variables.image_reader.pulse = value
        if self.variables.vcount == 0:
            self.variables.vmin = numpy.min(self.variables.image_reader.pulse_data[:, :])
            self.variables.vmax = numpy.max(self.variables.image_reader.pulse_data[:, :])
        elif self.variables.vcount < 5:
            self.variables.vmin = min(self.variables.vmin, numpy.min(self.variables.image_reader.pulse_data[:, :]))
            self.variables.vmax = max(self.variables.vmax, numpy.max(self.variables.image_reader.pulse_data[:, :]))
        if self.variables.vmin != self.variables.vmax:
            self.variables.vcount += 1

        self.slider.entry_pulse.set_text(str(value))
        self.slider.var_pulse_number.set(value)
        self.display_in_pyplot_frame()

    def exit(self):
        self.root.destroy()

    def _refresh_vdata(self):
        self.variables.vmin = 0
        self.variables.vmax = 0
        self.variables.vcount = 0

    # noinspection PyUnusedLocal
    def handle_remap_change(self, event):
        """
        Handle that the remap for the image canvas has changed.

        Parameters
        ----------
        event
        """

        pass

    # noinspection PyUnusedLocal
    def handle_image_index_changed(self, event):
        """
        Handle that the image index has changed.

        Parameters
        ----------
        event
        """
        if self.variables.image_reader is None:
            return

        self._refresh_vdata()
        self.variables.animating = False
        self.variables.image_reader.index = self.slider.cbx_channel.current()
        self.my_populate_metaicon()
        self.slider.cbx_channel.selection_clear()
        # Update number of pulses
        pulse_count = self.variables.image_reader.pulse_count
        self.slider.fullscale.configure(text=str(pulse_count-1))
        self.slider.scale_pulse.configure(to=pulse_count-1)
        self.set_pulse(0, override=True)

    def handle_pulse_changed_entry(self, event):
        if self.variables.image_reader is None:
            return

        if self.variables.animating:
            self.variables.animating = False

        self.set_pulse(int(self.slider.entry_pulse.get()))

    def handle_pulse_changed_slider(self, event):
        if self.variables.image_reader is None:
            return

        if self.variables.animating:
            self.variables.animating = False

        self.set_pulse(int(float(self.slider.var_pulse_number.get())))

    def pulse_step(self, direction):
        if self.variables.image_reader is None:
            return

        current_pulse = int(float(self.slider.var_pulse_number.get()))
        new_pulse = current_pulse + direction
        # todo: should we roll over by default?
        if new_pulse < 0:
            new_pulse += self.variables.image_reader.pulse_count
        elif new_pulse >= self.variables.image_reader.pulse_count:
            new_pulse -= self.variables.image_reader.pulse_count
        self.set_pulse(new_pulse)

    def pulse_step_prev(self):
        if self.variables.animating:
            self.variables.animating = False
        self.pulse_step(-1)

    def pulse_animate_backwards(self):
        if self.variables.animating:
            self.variables.animating = False
            return

        self.variables.animating = True
        self._pulse_animate(-1)

    def pulse_step_next(self):
        if self.variables.animating:
            self.variables.animating = False

        self.pulse_step(+1)

    def pulse_animate_forward(self):
        if self.variables.animating:
            self.variables.animating = False
            return

        self.variables.animating = True
        self._pulse_animate(1)

    def _pulse_animate(self, direction):
        if not self.variables.animating:
            return

        self.pulse_step(direction)
        self.after(self.variables.animation_delay, self._pulse_animate, direction)

    def update_reader(self, the_reader, update_browse=None):
        """
        Update the reader.

        Parameters
        ----------
        the_reader : None|str|CRSDTypeReader|STFTCanvasImageReader
        update_browse : None|str
        """

        if the_reader is None:
            return

        if update_browse is not None:
            self.variables.browse_directory = update_browse
        elif isinstance(the_reader, str):
            self.variables.browse_directory = os.path.split(the_reader)[0]

        if isinstance(the_reader, str):
            the_reader = STFTCanvasImageReader(the_reader)

        if isinstance(the_reader, CRSDTypeReader):
            the_reader = STFTCanvasImageReader(the_reader)

        if not isinstance(the_reader, STFTCanvasImageReader):
            raise TypeError('Got unexpected input for the reader')

        self._refresh_vdata()
        self.variables.animating = False
        self.variables.image_reader = the_reader
        self.display_in_pyplot_frame()
        self.set_title()
        self.my_populate_metaicon()
        self.my_populate_metaviewer()
        identifiers = [
            entry.Identifier for entry in the_reader.base_reader.crsd_meta.Channel.Parameters]
        self.update_combobox(identifiers)  # which one is set here...

        self.slider.fullscale['text'] = str(self.variables.image_reader.pulse_count-1)
        self.slider.scale_pulse['to'] = self.variables.image_reader.pulse_count-1
        self.slider.var_pulse_number.set(0)

    def callback_select_files(self):
        fname = askopenfilename(initialdir=self.variables.browse_directory,
                                filetypes=[crsd_files, all_files])
        if fname is None or fname in ['', ()]:
            return

        the_reader = STFTCanvasImageReader(fname)
        self.update_reader(the_reader, update_browse=os.path.split(fname)[0])

    def display_in_pyplot_frame(self):
        times = 1e6*self.variables.image_reader.times
        frequencies = 1e-9*self.variables.image_reader.frequencies
        image_data = self.variables.image_reader.pulse_data[:, :]
        image_data = image_data[::-1, :]
        if self.variables.vcount > 0:
            self.pyplot_panel.update_image(
                image_data, aspect='auto', vmin=self.variables.vmin, vmax=self.variables.vmax,
                extent=[times[0], times[-1], frequencies[0], frequencies[-1]])
        else:
            self.pyplot_panel.update_image(
                image_data, aspect='auto',
                extent=[times[0], times[-1], frequencies[0], frequencies[-1]])

    def update_combobox(self, identifiers):
        # Update channels combobox.
        # Get channel identifiers from metadata.
        self.slider.cbx_channel['values'] = identifiers
        self.slider.cbx_channel.set(identifiers[0])
        self.slider.cbx_channel.configure(state='readonly')

    def my_populate_metaicon(self):
        """
        Populate the metaicon.
        """

        self.populate_metaicon(self.variables.image_reader)

    def my_populate_metaviewer(self):
        """
        Populate the metaviewer.
        """

        self.populate_metaviewer(self.variables.image_reader)


def main(reader=None):
    """
    Main method for initializing the pulse explorer

    Parameters
    ----------
    reader : None|str|CRSDTypeReader|STFTCanvasImageReader
    """

    root = tkinter.Tk()

    the_style = ttk.Style()
    the_style.theme_use('default')

    app = PulseExplorer(root, reader)

    if reader is not None:
        app.update_reader(reader)

    root.mainloop()


if __name__ == '__main__':
    import argparse

    parser = argparse.ArgumentParser(
        description="Open the pulse explorer with optional input file.",
        formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument(
        '-i', '--input', metavar='input', default=None,
        help='The path to the optional image file for opening.')
    args = parser.parse_args()

    main(reader=args.input)



"""
Helper classes mapping the sarpy readers into the CanvasImageReader pattern.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


import logging
import numpy
from typing import List, Tuple
import gc

from tk_builder.image_reader import CanvasImageReader

from sarpy.io.general.base import AbstractReader, SarpyIOError
from sarpy.visualization.remap import get_remap_list, get_registered_remap, RemapFunction

from sarpy.io.complex.converter import open_complex
from sarpy.io.complex.base import SICDTypeReader
from sarpy.io.complex.aggregate import AggregateComplexReader
from sarpy.io.complex.sicd_elements.SICD import SICDType

from sarpy.io.product.converter import open_product
from sarpy.io.product.base import SIDDTypeReader
from sarpy.io.product.sidd1_elements.SIDD import SIDDType as SIDDType1
from sarpy.io.product.sidd2_elements.SIDD import SIDDType as SIDDType2


from sarpy.io.phase_history.converter import open_phase_history
from sarpy.io.phase_history.base import CPHDTypeReader
from sarpy.io.phase_history.cphd1_elements.CPHD import CPHDType as CPHDType1
from sarpy.io.phase_history.cphd0_3_elements.CPHD import CPHDType as CPHDType0_3

from sarpy.io.received.converter import open_received
from sarpy.io.received.base import CRSDTypeReader
from sarpy.io.received.crsd1_elements.CRSD import CRSDType

from sarpy.io.general.converter import open_general


def _get_default_remap():
    """
    Gets the default remap function.

    Returns
    -------
    RemapFunction
    """

    return get_remap_list()[0][1]


#######
# general reader

class GeneralCanvasImageReader(CanvasImageReader):
    """
    This is a general image reader of unknown type. There may be trouble
    with the image segments of unexpected type.
    """

    __slots__ = ('_base_reader', '_chippers', '_index', '_data_size')

    def __init__(self, reader):
        """

        Parameters
        ----------
        reader : str|AbstractReader
            The reader, or path to appropriate data file.
        """

        # initialize
        self._base_reader = None
        self._chippers = None
        self._data_size = None
        self._index = None
        # set the reader
        self.base_reader = reader

    @property
    def base_reader(self):
        # type: () -> AbstractReader
        """
        AbstractReader: The reader object
        """

        return self._base_reader

    @base_reader.setter
    def base_reader(self, value):
        if isinstance(value, str):
            value = open_general(value)
        if not isinstance(value, AbstractReader):
            raise TypeError('base_reader must be of type AbstractReader, got type {}'.format(type(value)))
        self._base_reader = value
        # noinspection PyProtectedMember
        self._chippers = value._get_chippers_as_tuple()
        self.index = 0

    @property
    def reader_type(self):
        """
        str: The reader type.
        """

        return self.base_reader.reader_type

    @property
    def index(self):
        """
        int: The reader index.
        """

        return self._index

    @index.setter
    def index(self, value):
        value = int(value)
        data_sizes = self.base_reader.get_data_size_as_tuple()
        if not (0 <= value < len(data_sizes)):
            logging.error(
                'The index property must be 0 <= index < {}, '
                'and got argument {}. Setting to 0.'.format(len(data_sizes), value))
            value = 0
        self._index = value
        self._data_size = data_sizes[value]

    @property
    def file_name(self):
        return None if self.base_reader is None else self.base_reader.file_name

    @property
    def remapable(self):
        return False

    @property
    def remap_function(self):
        return None

    @property
    def image_count(self):
        return 0 if self._chippers is None else len(self._chippers)

    def get_meta_data(self):
        """
        Gets one of a varieties of metadata structure.

        Returns
        -------
        Any
        """

        if isinstance(self.base_reader, SICDTypeReader):
            if self._index is None:
                return None
            # noinspection PyUnresolvedReferences
            return self.base_reader.get_sicds_as_tuple()[self._index]
        elif isinstance(self.base_reader, SIDDTypeReader):
            if self._index is None:
                return None
            # noinspection PyUnresolvedReferences
            return self.base_reader.get_sidds_as_tuple()[self._index]
        elif isinstance(self.base_reader, CPHDTypeReader):
            # noinspection PyUnresolvedReferences
            return self.base_reader.cphd_meta
        elif isinstance(self.base_reader, CRSDTypeReader):
            # noinspection PyUnresolvedReferences
            return self.base_reader.crsd_meta
        return None

    def __getitem__(self, item):
        return self._chippers[self.index].__getitem__(item)

    def __del__(self):
        # noinspection PyBroadException
        try:
            del self._chippers
            gc.collect()
        except Exception:
            pass


########
# general complex type reader structure - really just sets the remap

class ComplexCanvasImageReader(GeneralCanvasImageReader):
    __slots__ = ('_base_reader', '_chippers', '_index', '_data_size', '_remap_function')

    def __init__(self, reader):
        """

        Parameters
        ----------
        reader : str|AbstractReader
            The complex valued reader, or path to appropriate data file.
        """

        self._remap_function = _get_default_remap()
        GeneralCanvasImageReader.__init__(self, reader)

    @property
    def base_reader(self):
        # type: () -> AbstractReader
        """
        AbstractReader: The complex-valued reader object
        """

        return self._base_reader

    @base_reader.setter
    def base_reader(self, value):
        if isinstance(value, str):
            reader = None

            # try to open as sicd type
            try:
                reader = open_complex(value)
            except SarpyIOError:
                pass

            # try to open as phase_history
            if reader is None:
                try:
                    reader = open_phase_history(value)
                except SarpyIOError:
                    pass

            if reader is None:
                try:
                    reader = open_received(value)
                except SarpyIOError:
                    pass

            if reader is None:
                raise SarpyIOError('Could not open file {} as a one of the complex type readers'.format(value))
            value = reader
        elif isinstance(value, (tuple, list)):
            value = AggregateComplexReader(value)

        if not isinstance(value, AbstractReader):
            raise TypeError('base_reader must be of type AbstractReader, got type {}'.format(type(value)))
        if value.reader_type not in ["SICD", "CPHD", "CRSD"]:
            raise SarpyIOError(
                'base_reader.reader_type must be "SICD", "CPHD", or "CRSD", got {}'.format(value.reader_type))
        self._base_reader = value
        # noinspection PyProtectedMember
        self._chippers = value._get_chippers_as_tuple()
        self._index = 0
        self._data_size = value.get_data_size_as_tuple()[0]

    @property
    def remapable(self):
        return True

    @property
    def remap_function(self):
        return self._remap_function

    def __getitem__(self, item):
        cdata = self._chippers[self.index].__getitem__(item)
        return self.remap_complex_data(cdata)

    def remap_complex_data(self, complex_data):
        """
        Perform the remap on the complex data.

        Parameters
        ----------
        complex_data : numpy.ndarray

        Returns
        -------
        numpy.ndarray
        """

        return self._remap_function(complex_data)

    def set_remap_type(self, remap_type):
        if callable(remap_type):
            self._remap_function = remap_type
        elif isinstance(remap_type, str):
            self._remap_function = get_registered_remap(remap_type, _get_default_remap())
        else:
            default_remap = _get_default_remap()
            logging.error(
                'Got unexpected value for remap `{}`, using `{}`'.format(remap_type, default_remap.name))
            self._remap_function = default_remap


########
# SICD specific type readers

class SICDTypeCanvasImageReader(ComplexCanvasImageReader):

    def __init__(self, reader):
        """

        Parameters
        ----------
        reader : str|SICDTypeReader
            The sicd type reader, or path to appropriate data file.
        """

        ComplexCanvasImageReader.__init__(self, reader)

    @property
    def base_reader(self):
        # type: () -> SICDTypeReader
        """
        SICDTypeReader: The complex-valued reader object
        """

        return self._base_reader

    @base_reader.setter
    def base_reader(self, value):
        if isinstance(value, str):
            reader = None
            try:
                reader = open_complex(value)
            except SarpyIOError:
                pass

            if reader is None:
                raise SarpyIOError('Could not open file {} as a SICD type reader'.format(value))
            value = reader
        elif isinstance(value, (tuple, list)):
            value = AggregateComplexReader(value)

        if not isinstance(value, SICDTypeReader):
            raise TypeError('base_reader must be a SICDTypeReader, got type {}'.format(type(value)))
        self._base_reader = value
        # noinspection PyProtectedMember
        self._chippers = value._get_chippers_as_tuple()
        self._index = 0
        self._data_size = value.get_data_size_as_tuple()[0]

    def get_sicd(self):
        """
        Gets the relevant SICD structure.

        Returns
        -------
        None|SICDType
        """

        if self._index is None:
            return None
        return self.base_reader.get_sicds_as_tuple()[self._index]

    def transform_coordinates(self, image_coordinates):
        sicd = self.get_sicd()
        if sicd is None:
            return None, 'NONE'

        return sicd.project_image_to_ground_geo(image_coordinates, projection_type='HAE'), 'LLH_HAE'


class QuadPolCanvasImageReader(ComplexCanvasImageReader):
    __slots__ = (
        '_base_reader', '_chippers', '_sicd_partitions', '_index', '_index_ordering',
        '_data_size', '_remap_function')

    def __init__(self, reader):
        """

        Parameters
        ----------
        reader : str|SICDTypeReader
            The sicd type reader, or path to appropriate data file.
        """

        ComplexCanvasImageReader.__init__(self, reader)

    @property
    def base_reader(self):
        # type: () -> SICDTypeReader
        """
        SICDTypeReader: The complex-valued reader object
        """

        return self._base_reader

    @base_reader.setter
    def base_reader(self, value):
        if isinstance(value, str):
            value = open_complex(value)
        elif isinstance(value, (list, tuple)):
            value = AggregateComplexReader(value)
        if not isinstance(value, SICDTypeReader):
            raise TypeError('Requires that the input is a sicd type reader object. Got type {}'.format(type(value)))

        self._base_reader = value
        # noinspection PyProtectedMember
        self._chippers = value._get_chippers_as_tuple()
        self._sicd_partitions = value.get_sicd_partitions()
        try:
            self.index = 0
        except ValueError:
            # the index will be uninitialized
            pass

    @property
    def sicd_partition(self):
        """
        Tuple[Tuple[int]]: The partitioning of the reader indices into matching components.
        """

        return self._sicd_partitions

    @property
    def index(self):
        """
        int: the SICD partition index.
        """

        return self._index

    @index.setter
    def index(self, value):
        if self._sicd_partitions is None:
            return

        value = int(value)
        if not (0 <= value < len(self.sicd_partition)):
            raise ValueError('index must be on the range 0 <= index < {}'.format(len(self.sicd_partition)))
        indices = self.sicd_partition[value]
        # determine appropriate ordering for index collection
        data_sizes = self.base_reader.get_data_size_as_tuple()
        data_size = data_sizes[indices[0]]
        for entry in indices:
            if data_sizes[entry] != data_size:
                raise ValueError(
                    'All entries of sicd partition {} do not yield sicd elements of '
                    'the same row/column size.'.format(value))
        self._index = value
        self._data_size = data_size
        self._order_indices(value, indices)

    def _order_indices(self, index, indices):
        """
        Determine the appropriate ordering for indices.

        Parameters
        ----------
        index : int
        indices : Tuple[int]

        Returns
        -------
        None
        """

        def revert():
            self._index = None
            self._index_ordering = None
            self._data_size = None

        if len(indices) == 1:
            self._index_ordering = indices
            return

        sicds = self.base_reader.get_sicds_as_tuple()
        our_sicds = [sicds[entry] for entry in indices]  # type: List[SICDType]
        pols = [entry.ImageFormation.TxRcvPolarizationProc for entry in our_sicds]
        if len(indices) == 2:
            pols_set = set(pols)
            if len(pols_set) != 2:
                ordered_indices = None
            else:
                ordered_indices = None
                for desired_order in [['CV', 'CH'], ['VV', 'HH']]:
                    if pols_set == set(desired_order):
                        ordered_indices = [pols.index(entry) for entry in desired_order]
                        break
        elif len(indices) == 4:
            pols_set = set(pols)
            if len(pols_set) != 4:
                ordered_indices = None
            else:
                ordered_indices = None
                for desired_order in [['VV', 'VH', 'HV', 'HH'], ]:
                    if pols_set == set(desired_order):
                        ordered_indices = [pols.index(entry) for entry in desired_order]
                        break
        else:
            ordered_indices = None

        if ordered_indices is None:
            revert()
            raise ValueError(
                'Got unhandled polarization states for partition {}'.format(pols, index))
        self._index_ordering = ordered_indices

    def __getitem__(self, item):
        def get_cdata(the_index):
            return self._chippers[the_index].__getitem__(item)

        if self._index_ordering is None:
            return None
        if len(self._index_ordering) == 1:
            complex_data = get_cdata(self._index_ordering[0])
            return self._remap_function(complex_data)

        complex_data = [get_cdata(entry) for entry in self._index_ordering]
        out_size = complex_data[0].shape
        for entry in complex_data:
            if entry.shape != out_size:
                raise ValueError('Got unexpected mismatch in sizes {} and {}'.format(entry.shape, out_size))
        data_mean = float(max(numpy.mean(numpy.abs(entry)) for entry in complex_data))
        rgb_image = numpy.zeros(out_size + (3, ), dtype='uint8')
        if len(self._index_ordering) == 2:
            try:
                rgb_image[:, :, 0] = self._remap_function(complex_data[0], data_mean=data_mean)
                rgb_image[:, :, 2] = self._remap_function(complex_data[1], data_mean=data_mean)
            except Exception:
                rgb_image[:, :, 0] = self._remap_function(complex_data[0])
                rgb_image[:, :, 2] = self._remap_function(complex_data[1])
        elif len(self._index_ordering) == 4:
            try:
                rgb_image[:, :, 0] = self._remap_function(complex_data[0], data_mean=data_mean)
                rgb_image[:, :, 1] = self._remap_function(complex_data[1], data_mean=data_mean)/2 + \
                                     self._remap_function(complex_data[2], data_mean=data_mean)/2
                rgb_image[:, :, 2] = self._remap_function(complex_data[3], data_mean=data_mean)
            except Exception:
                rgb_image[:, :, 0] = self._remap_function(complex_data[0])
                rgb_image[:, :, 1] = self._remap_function(complex_data[1])/2 + \
                                     self._remap_function(complex_data[2])/2
                rgb_image[:, :, 2] = self._remap_function(complex_data[3])
        else:
            raise ValueError('Got unhandled case for collection {}'.format(self._index_ordering))
        return rgb_image

    def get_sicd(self):
        """
        Gets the relevant SICD structure.

        Returns
        -------
        None|SICDType
        """

        if self._index is None:
            return None
        return self.base_reader.get_sicds_as_tuple()[self._index]

    def transform_coordinates(self, image_coordinates):
        sicd = self.get_sicd()
        if sicd is None:
            return None, 'NONE'

        return sicd.project_image_to_ground_geo(image_coordinates, projection_type='HAE'), 'LLH_HAE'


#######
# Phase history specific type reader

class CPHDTypeCanvasImageReader(ComplexCanvasImageReader):

    def __init__(self, reader):
        """

        Parameters
        ----------
        reader : str|CPHDTypeReader
            The cphd type reader, or path to appropriate data file.
        """

        ComplexCanvasImageReader.__init__(self, reader)

    @property
    def base_reader(self):
        # type: () -> CPHDTypeReader
        """
        CPHDTypeReader: The cphd reader object
        """

        return self._base_reader

    @base_reader.setter
    def base_reader(self, value):
        if isinstance(value, str):
            reader = None
            try:
                reader = open_phase_history(value)
            except SarpyIOError:
                pass

            if reader is None:
                raise SarpyIOError('Could not open file {} as a CPHD reader'.format(value))
            value = reader

        if not isinstance(value, CPHDTypeReader):
            raise TypeError('base_reader must be a CPHDReader, got type {}'.format(type(value)))
        self._base_reader = value
        # noinspection PyProtectedMember
        self._chippers = value._get_chippers_as_tuple()
        self._index = 0
        self._data_size = value.get_data_size_as_tuple()[0]

    def get_cphd(self):
        """
        Gets the relevant CPHD structure.

        Returns
        -------
        None|CPHDType1|CPHDType0_3
        """

        return self.base_reader.cphd_meta


#######
# Received data specific type reader

class CRSDTypeCanvasImageReader(ComplexCanvasImageReader):

    def __init__(self, reader):
        """

        Parameters
        ----------
        reader : str|CRSDTypeReader
            The crsd type reader, or path to appropriate data file.
        """

        ComplexCanvasImageReader.__init__(self, reader)

    @property
    def base_reader(self):
        # type: () -> CRSDTypeReader
        """
        CRSDTypeReader: The crsd reader object
        """

        return self._base_reader

    @base_reader.setter
    def base_reader(self, value):
        if isinstance(value, str):
            reader = None
            try:
                reader = open_received(value)
            except SarpyIOError:
                pass

            if reader is None:
                raise SarpyIOError('Could not open file {} as a CRSD reader'.format(value))
            value = reader

        if not isinstance(value, CRSDTypeReader):
            raise TypeError('base_reader must be a CRSDReader, got type {}'.format(type(value)))
        self._base_reader = value
        # noinspection PyProtectedMember
        self._chippers = value._get_chippers_as_tuple()
        self._index = 0
        self._data_size = value.get_data_size_as_tuple()[0]

    def get_crsd(self):
        """
        Gets the relevant CRSD structure.

        Returns
        -------
        None|CRSDType
        """

        return self.base_reader.crsd_meta


######
# SIDD specific type reader

class DerivedCanvasImageReader(GeneralCanvasImageReader):
    __slots__ = ('_base_reader', '_chippers', '_index', '_data_size')

    def __init__(self, reader):
        """

        Parameters
        ----------
        reader : str|SIDDTypeReader
            The sidd type reader, or path to appropriate data file.
        """

        GeneralCanvasImageReader.__init__(self, reader)

    @property
    def base_reader(self):
        # type: () -> SIDDTypeReader
        """
        SIDDTypeReader: The SIDD based reader object
        """

        return self._base_reader

    @base_reader.setter
    def base_reader(self, value):
        if isinstance(value, str):
            value = open_product(value)
        if not isinstance(value, SIDDTypeReader):
            raise TypeError('base_reader must be a SIDDTypeReader, got type {}'.format(type(value)))
        self._base_reader = value
        # noinspection PyProtectedMember
        self._chippers = value._get_chippers_as_tuple()
        self._index = 0
        self._data_size = value.get_data_size_as_tuple()[0]

    @property
    def remapable(self):
        return False

    @property
    def remap_function(self):
        return None

    def get_sidd(self):
        """
        Gets the relevant SIDD structure.

        Returns
        -------
        None|SIDDType1|SIDDType2
        """

        if self._index is None:
            return None
        return self.base_reader.get_sidds_as_tuple()[self._index]

    def transform_coordinates(self, image_coordinates):
        sidd = self.get_sidd()
        if sidd is None:
            return None, 'NONE'

        return sidd.project_image_to_ground_geo(image_coordinates, projection_type='HAE'), 'LLH_HAE'



# -*- coding: utf-8 -*-
"""
This module provides a general SAR image viewer tool.
"""

__classification__ = "UNCLASSIFIED"
__author__ = ("Jason Casey", "Thomas McCullough")

import os

import tkinter
from tkinter import ttk
from tkinter.filedialog import askopenfilenames, askdirectory

from tk_builder.base_elements import StringDescriptor, TypedDescriptor
from tk_builder.image_reader import CanvasImageReader
from tk_builder.panels.pyplot_image_panel import PyplotImagePanel
from tk_builder.panels.image_panel import ImagePanel
from tk_builder.widgets.basic_widgets import Frame

from sarpy_apps.supporting_classes.file_filters import common_use_collection
from sarpy_apps.supporting_classes.image_reader import SICDTypeCanvasImageReader, \
    DerivedCanvasImageReader, CPHDTypeCanvasImageReader, CRSDTypeCanvasImageReader, \
    GeneralCanvasImageReader
from sarpy_apps.supporting_classes.widget_with_metadata import WidgetWithMetadata

from sarpy.io.general.base import BaseReader
from sarpy.io.complex.base import SICDTypeReader
from sarpy.io.product.base import SIDDTypeReader
from sarpy.io.phase_history.base import CPHDTypeReader
from sarpy.io.received.base import CRSDTypeReader
from sarpy.io import open as open_general


class AppVariables(object):
    browse_directory = StringDescriptor(
        'browse_directory', default_value=os.path.expanduser('~'),
        docstring='The directory for browsing for file selection.')  # type: str
    image_reader = TypedDescriptor(
        'image_reader', CanvasImageReader, docstring='')  # type: CanvasImageReader


class ImageViewer(Frame, WidgetWithMetadata):
    def __init__(self, primary, reader=None, **kwargs):
        """

        Parameters
        ----------
        primary : tkinter.Toplevel|tkinter.Tk
        reader
        kwargs
        """

        self.root = primary
        Frame.__init__(self, primary, **kwargs)
        WidgetWithMetadata.__init__(self, primary)
        self.pack(fill=tkinter.BOTH, expand=tkinter.YES)
        self.primary = tkinter.PanedWindow(self, sashrelief=tkinter.RIDGE, orient=tkinter.HORIZONTAL)

        self.variables = AppVariables()

        self.image_panel = ImagePanel(self.primary, borderwidth=0)  # type: ImagePanel
        self.primary.add(
            self.image_panel, width=400, height=700, padx=5, pady=5, sticky=tkinter.NSEW,
            stretch=tkinter.FIRST)

        self.pyplot_panel = PyplotImagePanel(self.primary)  # type: PyplotImagePanel
        self.primary.add(
            self.pyplot_panel, width=400, height=700, padx=5, pady=5, sticky=tkinter.NSEW)

        self.primary.pack(fill=tkinter.BOTH, expand=tkinter.YES)

        self.set_title()

        # define menus
        menubar = tkinter.Menu()
        # file menu
        filemenu = tkinter.Menu(menubar, tearoff=0)
        filemenu.add_command(label="Open Image", command=self.callback_select_files)
        filemenu.add_command(label="Open Directory", command=self.callback_select_directory)
        filemenu.add_separator()
        filemenu.add_command(label="Exit", command=self.exit)
        # menus for informational popups
        popups_menu = tkinter.Menu(menubar, tearoff=0)
        popups_menu.add_command(label="Metaicon", command=self.metaicon_popup)
        popups_menu.add_command(label="Metaviewer", command=self.metaviewer_popup)
        self._valid_data_shown = tkinter.IntVar(self, value=0)
        popups_menu.add_checkbutton(
            label='ValidData', variable=self._valid_data_shown, command=self.show_valid_data)
        # ensure menus cascade
        menubar.add_cascade(label="File", menu=filemenu)
        menubar.add_cascade(label="Metadata", menu=popups_menu)

        primary.config(menu=menubar)

        # hide extraneous tool elements
        self.image_panel.hide_tools('shape_drawing')
        self.image_panel.hide_shapes()

        # bind canvas events for proper functionality
        # this makes for bad performance on a larger image - do not activate
        # self.image_panel.canvas.bind('<<SelectionChanged>>', self.handle_selection_change)
        self.image_panel.canvas.bind('<<SelectionFinalized>>', self.handle_selection_change)
        self.image_panel.canvas.bind('<<RemapChanged>>', self.handle_remap_change)
        self.image_panel.canvas.bind('<<ImageIndexChanged>>', self.handle_image_index_changed)

        self.update_reader(reader, update_browse=None)

    def set_title(self):
        """
        Sets the window title.
        """

        file_name = None if self.variables.image_reader is None else self.variables.image_reader.file_name
        if file_name is None:
            the_title = "Image Viewer"
        elif isinstance(file_name, (list, tuple)):
            the_title = "Image Viewer, Multiple Files"
        else:
            the_title = "Image Viewer for {}".format(os.path.split(file_name)[1])
        self.winfo_toplevel().title(the_title)

    def exit(self):
        self.primary.destroy()
        self.root.destroy()

    def show_valid_data(self):
        if self.variables.image_reader is None or \
                not isinstance(self.variables.image_reader, SICDTypeCanvasImageReader):
            return

        the_value = self._valid_data_shown.get()
        if the_value == 1:
            # we just checked on
            sicd = self.variables.image_reader.get_sicd()
            if sicd.ImageData.ValidData is not None:
                self.image_panel.canvas.show_valid_data(sicd.ImageData.ValidData.get_array(dtype='float64'))
        else:
            # we checked it off
            try:
                valid_data_id = self.image_panel.canvas.variables.get_tool_shape_id_by_name('VALID_DATA')
                self.image_panel.canvas.hide_shape(valid_data_id)
            except KeyError:
                pass

    # noinspection PyUnusedLocal
    def handle_selection_change(self, event):
        """
        Handle a change in the selection area.

        Parameters
        ----------
        event
        """

        if self.variables.image_reader is None:
            return

        full_image_width = self.image_panel.canvas.variables.state.canvas_width
        fill_image_height = self.image_panel.canvas.variables.state.canvas_height
        self.image_panel.canvas.zoom_to_canvas_selection((0, 0, full_image_width, fill_image_height))
        self.display_canvas_rect_selection_in_pyplot_frame()

    # noinspection PyUnusedLocal
    def handle_remap_change(self, event):
        """
        Handle that the remap for the image canvas has changed.

        Parameters
        ----------
        event
        """
        if self.variables.image_reader is not None:
            self.display_canvas_rect_selection_in_pyplot_frame()

    # noinspection PyUnusedLocal
    def handle_image_index_changed(self, event):
        """
        Handle that the image index has changed.

        Parameters
        ----------
        event
        """

        self.my_populate_metaicon()
        self.show_valid_data()

    def update_reader(self, the_reader, update_browse=None):
        """
        Update the reader.

        Parameters
        ----------
        the_reader : None|str|BaseReader|CanvasImageReader
        update_browse : None|str
        """

        if the_reader is None:
            return

        if update_browse is not None:
            self.variables.browse_directory = update_browse
        elif isinstance(the_reader, str):
            self.variables.browse_directory = os.path.split(the_reader)[0]

        if isinstance(the_reader, str):
            the_reader = open_general(the_reader)

        if isinstance(the_reader, SICDTypeReader):
            the_reader = SICDTypeCanvasImageReader(the_reader)
        elif isinstance(the_reader, SIDDTypeReader):
            the_reader = DerivedCanvasImageReader(the_reader)
        elif isinstance(the_reader, CPHDTypeReader):
            the_reader = CPHDTypeCanvasImageReader(the_reader)
        elif isinstance(the_reader, CRSDTypeReader):
            the_reader = CRSDTypeCanvasImageReader(the_reader)
        elif isinstance(the_reader, BaseReader):
            the_reader = GeneralCanvasImageReader(the_reader)

        if not isinstance(the_reader, CanvasImageReader):
            raise TypeError('Got unexpected input for the reader')

        # change the tool to view
        self.image_panel.canvas.current_tool = 'VIEW'
        self.image_panel.canvas.current_tool = 'VIEW'
        # update the reader
        self.variables.image_reader = the_reader
        self.image_panel.set_image_reader(the_reader)
        self.set_title()
        # refresh appropriate GUI elements
        self.pyplot_panel.make_blank()
        self.my_populate_metaicon()
        self.my_populate_metaviewer()
        self.show_valid_data()

    def callback_select_files(self):
        fnames = askopenfilenames(initialdir=self.variables.browse_directory, filetypes=common_use_collection)
        if fnames is None or fnames in ['', ()]:
            return

        if len(fnames) > 1:
            the_reader = SICDTypeCanvasImageReader(fnames)
            self.update_reader(the_reader, update_browse=os.path.split(fnames[0])[0])
        else:
            self.update_reader(fnames[0], update_browse=os.path.split(fnames[0])[0])

    def callback_select_directory(self):
        dirname = askdirectory(initialdir=self.variables.browse_directory, mustexist=True)
        if dirname is None or dirname in [(), '']:
            return
        # NB: handle non-complex data possibilities here?
        the_reader = SICDTypeCanvasImageReader(dirname)
        self.update_reader(the_reader, update_browse=os.path.split(dirname)[0])

    def display_canvas_rect_selection_in_pyplot_frame(self):
        def get_extent(coords):
            left = min(coords[1::2])
            right = max(coords[1::2])
            top = max(coords[0::2])
            bottom = min(coords[0::2])
            return left, right, top, bottom

        threshold = self.image_panel.canvas.variables.config.select_size_threshold

        select_id = self.image_panel.canvas.variables.get_tool_shape_id_by_name('SELECT')
        rect_coords = self.image_panel.canvas.get_shape_image_coords(select_id)
        extent = get_extent(rect_coords)

        if abs(extent[1] - extent[0]) < threshold or abs(extent[2] - extent[3]) < threshold:
            self.pyplot_panel.make_blank()
        else:
            image_data = self.image_panel.canvas.get_image_data_in_canvas_rect_by_id(select_id)
            if image_data is not None:
                self.pyplot_panel.update_image(image_data, extent=extent)
            else:
                self.pyplot_panel.make_blank()

    def my_populate_metaicon(self):
        """
        Populate the metaicon.
        """

        self.populate_metaicon(self.variables.image_reader)

    def my_populate_metaviewer(self):
        """
        Populate the metaviewer.
        """

        self.populate_metaviewer(self.variables.image_reader)


def main(reader=None):
    """
    Main method for initializing the aperture tool

    Parameters
    ----------
    reader : None|str|BaseReader|CanvasImageReader
    """

    root = tkinter.Tk()

    the_style = ttk.Style()
    the_style.theme_use('classic')

    app = ImageViewer(root, reader=reader)
    root.geometry("1000x800")

    root.mainloop()


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser(
        description="Open the image viewer with optional input file.",
        formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument(
        '-i', '--input', metavar='input', default=None, help='The path to the optional image file for opening.')
    args = parser.parse_args()

    main(reader=args.input)



"""
A structured labeling tool
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"


import os
from typing import Union
from datetime import datetime

import tkinter
from tkinter import ttk, PanedWindow

from tkinter.messagebox import showinfo
from tkinter.filedialog import askopenfilename, asksaveasfilename

from tk_builder.base_elements import TypedDescriptor
from tk_builder.panels.image_panel import ImagePanel
from tk_builder.widgets.basic_widgets import Frame, Button, Label, Combobox, Entry
from tk_builder.widgets.derived_widgets import TreeviewWithScrolling, TextWithScrolling

from sarpy_apps.apps.annotation_tool import AppVariables as AppVariables_Annotate, \
    NamePanel, AnnotateButtons, AnnotateTabControl, AnnotationPanel, \
    AnnotationCollectionViewer, AnnotationCollectionPanel, AnnotationTool
from sarpy_apps.apps.labeling_tool.schema_editor import SchemaEditor, select_schema_entry
from sarpy_apps.supporting_classes.file_filters import all_files, json_files
from sarpy_apps.supporting_classes.image_reader import GeneralCanvasImageReader
from sarpy_apps.supporting_classes.widget_with_metadata import WidgetWithMetadata

from sarpy.annotation.label import FileLabelCollection, LabelFeature, \
    LabelMetadataList, LabelMetadata, LabelSchema
from sarpy.io.general.base import BaseReader


def get_default_schema():
    return LabelSchema(labels={'0': 'unknown'})


class AppVariables(AppVariables_Annotate):
    file_annotation_collection = TypedDescriptor(
        'file_annotation_collection', FileLabelCollection,
        docstring='The file annotation collection.')  # type: FileLabelCollection

    @property
    def label_schema(self):
        """
        None|LabelSchema: The label schema.
        """

        return None if self.file_annotation_collection is None else \
            self.file_annotation_collection.label_schema

    def get_label(self, label_id):
        if self.file_annotation_collection is None or label_id is None:
            return None
        return self.file_annotation_collection.label_schema.labels[label_id]

    def get_current_annotation_object(self):
        """
        Gets the current annotation object

        Returns
        -------
        None|LabelFeature
        """

        if self._current_feature_id is None:
            return None
        return self.file_annotation_collection.annotations[self._current_feature_id]


###########
# elements for editing the label details

class LabelSpecificsPanel(Frame):
    """
    Edit/Display widget for LabelMetadata object
    """

    def __init__(self, master, app_variables, **kwargs):
        """

        Parameters
        ----------
        master
        app_variables : AppVariables
        kwargs
        """

        self.app_variables = app_variables
        self.current_metadata_list = None  # type: Union[None, LabelMetadataList]
        self.in_progess_metadata = None  # type: Union[None, LabelMetadata]
        self.has_confidence = False
        Frame.__init__(self, master, **kwargs)

        self.object_type_label = Label(self, text='Type:', relief=tkinter.RIDGE, width=15, padding=3)
        self.object_type_label.grid(row=0, column=0, sticky='NEW', padx=3, pady=3)
        self.object_type_button = Button(self, text='<Choose>')
        self.object_type_button.grid(row=0, column=1, sticky='NEW', padx=3, pady=3)

        self.comment_label = Label(self, text='Comment:', relief=tkinter.RIDGE, width=15, padding=3)
        self.comment_label.grid(row=1, column=0, sticky='NEW', padx=3, pady=3)
        self.comment_text = TextWithScrolling(self, height=5)
        self.comment_text.frame.grid(row=1, column=1, sticky='NSEW', padx=3, pady=3)

        self.confidence_label = Label(self, text='Confidence:', relief=tkinter.RIDGE, width=15, padding=3)
        self.confidence_label.grid(row=2, column=0, sticky='NEW', padx=3, pady=3)
        self.confidence_combo = Combobox(self)
        self.confidence_combo.grid(row=2, column=1, sticky='NEW', padx=3, pady=3)

        self.user_id_label = Label(self, text='User ID:', relief=tkinter.RIDGE, width=15, padding=3)
        self.user_id_label.grid(row=3, column=0, sticky='NEW', padx=3, pady=3)
        self.user_id_value = Entry(self, text='', relief=tkinter.RIDGE, width=15, state='disabled')
        self.user_id_value.grid(row=3, column=1, sticky='NEW', padx=3, pady=3)

        self.timestamp_label = Label(self, text='Timestamp:', relief=tkinter.RIDGE, width=15, padding=3)
        self.timestamp_label.grid(row=4, column=0, sticky='NEW', padx=3, pady=3)
        self.timestamp_value = Entry(self, text='', relief=tkinter.RIDGE, width=15, state='disabled')
        self.timestamp_value.grid(row=4, column=1, sticky='NEW', padx=3, pady=3)

        self.frame1 = Frame(self, borderwidth=1, relief=tkinter.RIDGE)
        self.cancel_button = Button(self.frame1, text='Cancel')
        self.cancel_button.pack(side=tkinter.RIGHT)
        self.submit_button = Button(self.frame1, text='Submit')
        self.submit_button.pack(side=tkinter.RIGHT)
        self.frame1.grid(row=5, column=0, columnspan=2, sticky='NSEW')
        self.grid_rowconfigure(2, weight=1)
        self.grid_columnconfigure(1, weight=1)

        self.object_type_button.config(command=self.callback_select_object_type)
        self.cancel_button.config(command=self.callback_cancel)

    def _populate_values_into_inprogress(self):
        msg = ''
        status = True
        if self.in_progess_metadata.label_id is None:
            msg += 'The label ID must be set\n'
            status = False
        comment = self.comment_text.get_value()
        self.in_progess_metadata.comment = None if comment == '' else comment

        if self.has_confidence:
            confidence = self.confidence_combo.get()
            self.in_progess_metadata.confidence = None if confidence == '' else confidence
        return status, msg

    def _fill_metadata(self, entry):
        """
        Populates the values from entry.

        Parameters
        ----------
        entry : None|LabelMetadata
        """

        if entry is None:
            self.object_type_button.set_text('')
            self.comment_text.set_value('')
            self.confidence_combo.set_text('')
            self.user_id_value.set_text('')
            self.timestamp_value.set_text('')
        else:
            self.object_type_button.set_text(
                '<Choose>' if entry.label_id is None else self.app_variables.get_label(entry.label_id))
            self.comment_text.set_value('' if entry.comment is None else entry.comment)
            self.confidence_combo.set_text('' if entry.confidence is None else str(entry.confidence))
            self.user_id_value.set_text(entry.user_id)
            self.timestamp_value.set_text(datetime.utcfromtimestamp(entry.timestamp).isoformat('T'))

    def set_entry(self, index):
        """
        Sets the entry to display

        Parameters
        ----------
        index : None|int
        """

        def disable():
            self.object_type_button.config(state='disabled')
            self.comment_text.config(state='disabled')
            self.confidence_combo.config(state='disabled')
            self.cancel_button.config(state='disabled')
            self.submit_button.config(state='disabled')

        def enable():
            self.object_type_button.config(state='normal')
            self.comment_text.config(state='normal')
            if self.has_confidence:
                self.confidence_combo.config(state='normal')
            self.cancel_button.config(state='normal')
            self.submit_button.config(state='normal')

        if index is not None:
            entry = self.current_metadata_list[index]
            self._fill_metadata(entry)
            disable()
            return

        if self.current_metadata_list is None:
            # nothing to be done
            self._fill_metadata(None)
            disable()
            return

        if len(self.current_metadata_list) == 0:
            self.in_progess_metadata = LabelMetadata()
        else:
            self.in_progess_metadata = self.current_metadata_list[0].replicate()
            self.in_progess_metadata.comment = None
        self._fill_metadata(self.in_progess_metadata)
        enable()

    def update_annotation(self):
        feature = self.app_variables.get_current_annotation_object()
        if feature is None:
            self.current_metadata_list = None
            self.set_entry(None)
        else:
            self.current_metadata_list = feature.properties.parameters
            if len(self.current_metadata_list) > 0:
                self.set_entry(0)
            else:
                self.set_entry(None)

    def update_annotation_collection(self):
        if self.app_variables.file_annotation_collection is not None:
            confidence_values = self.app_variables.file_annotation_collection.label_schema.confidence_values
            if confidence_values is None:
                self.has_confidence = False
                self.confidence_combo.update_combobox_values([])
            else:
                self.has_confidence = True
                self.confidence_combo.update_combobox_values(['{}'.format(entry) for entry in confidence_values])
        self.update_annotation()

    def callback_select_object_type(self):
        if self.app_variables is None or \
                self.app_variables.file_annotation_collection is None or \
                self.in_progess_metadata is None:
            return  # this would be broken state, and should not happen

        current_value = self.in_progess_metadata.label_id
        value = select_schema_entry(self.app_variables.label_schema, start_id=current_value)
        if value is None:
            return

        self.in_progess_metadata.label_id = value
        self.object_type_button.set_text(self.app_variables.get_label(value))

    def callback_cancel(self):
        self.in_progess_metadata = None
        self.update_annotation()

    def callback_submit(self):
        # NB: this should be called by the controlling parent for holistic state change everywhere
        status, msg = self._populate_values_into_inprogress()
        if not status:
            showinfo('Incomplete data population', message=msg)
            return status

        self.current_metadata_list.insert_new_element(self.in_progess_metadata)
        self.in_progess_metadata = None
        return status


class LabelDetailsPanel(Frame):
    def __init__(self, master, app_variables, **kwargs):
        """

        Parameters
        ----------
        master
        app_variables : AppVariables
        kwargs
        """

        self.app_variables = app_variables
        self.current_annotation = None  # type: Union[None, LabelFeature]
        Frame.__init__(self, master, **kwargs)

        self.update_label_button = Button(
            self, text='Update Label', command=self.callback_update_label)  # type: Button
        self.update_label_button.grid(row=0, column=0, columnspan=2, sticky='NE', padx=3, pady=3)

        self.viewer = TreeviewWithScrolling(
            self, columns=('confidence', ), selectmode=tkinter.BROWSE)  # type: TreeviewWithScrolling
        self.viewer.heading('#0', text='Label')
        self.viewer.heading('#1', text='Confidence')
        self.viewer.frame.grid(row=1, column=0, sticky='NSEW', padx=3, pady=3)
        # NB: reference the frame for packing, since it's already packed into a frame

        self.label_specifics = LabelSpecificsPanel(self, app_variables, border=1, relief=tkinter.RIDGE)
        self.label_specifics.grid(row=1, column=1, sticky='NSEW', padx=3, pady=3)
        self.grid_rowconfigure(1, weight=1)
        self.grid_columnconfigure(0, weight=1)
        self.grid_columnconfigure(1, weight=1)
        self.update_annotation()

        self.viewer.bind('<<TreeviewSelect>>', self.callback_label_selected_on_viewer)

    def _set_focus(self, uid):
        self.viewer.set_selection_with_expansion(uid)

    def _empty_entries(self):
        """
        Empty all entries - for the purpose of reinitializing.
        """

        self.viewer.delete(*self.viewer.get_children())

    def _fill_treeview(self):
        """
        Fills the treeview based on the current annotation.
        """

        if self.current_annotation is None or \
                self.current_annotation.properties.parameters is None or \
                len(self.current_annotation.properties.parameters) == 0:
            self._empty_entries()
            self.label_specifics.update_annotation()
            self.label_specifics.set_entry(None)
            return

        self._empty_entries()
        label_list = self.current_annotation.properties.parameters
        for i, entry in enumerate(label_list):
            label = self.app_variables.get_label(entry.label_id)
            conf = '' if entry.confidence is None else entry.confidence
            self.viewer.insert('', 'end', iid=str(i), text=label, values=(conf, ))
        if len(label_list) > 0:
            self._set_focus('0')

    # noinspection PyUnusedLocal
    def callback_label_selected_on_viewer(self, event):
        label_index = self.viewer.focus()
        if label_index == '':
            return
        self.label_specifics.set_entry(int(label_index))

    def callback_update_label(self):
        self.label_specifics.set_entry(None)

    def set_annotation_feature(self, annotation_feature):
        self.current_annotation = annotation_feature
        self._fill_treeview()
        self.label_specifics.update_annotation()

    def update_annotation(self):
        annotation_feature = self.app_variables.get_current_annotation_object()
        self.set_annotation_feature(annotation_feature)

    def cancel(self):
        self._fill_treeview()  # probably unnecessary

    def save(self):
        self._fill_treeview()  # probably unnecessary


#############
# refactoring from annotation to accommodate minor differences in the basic elements

class LabelTabControl(AnnotateTabControl):
    def __init__(self, master, app_variables, **kwargs):
        AnnotateTabControl.__init__(self, master, app_variables, **kwargs)
        self.label_tab = LabelDetailsPanel(self, app_variables)
        self.tab_control.add(self.label_tab, text='Label')

    def update_annotation(self):
        self.details_tab.update_annotation()
        self.geometry_tab.update_annotation()
        self.label_tab.update_annotation()

    def _set_active_shapes(self):
        label_schema = self.app_variables.label_schema  # type: LabelSchema
        if label_schema is None or \
                label_schema.permitted_geometries is None or \
                len(label_schema.permitted_geometries) == 0:
            shapes = None
        else:
            shapes = []
            if 'point' in label_schema.permitted_geometries:
                shapes.append('point')
            if 'line' in label_schema.permitted_geometries:
                shapes.append('line')
            if 'polygon' in label_schema.permitted_geometries:
                shapes.extend(['rectangle', 'ellipse', 'polygon'])
        self.geometry_tab.geometry_buttons.set_active_shapes(shapes)

    def update_annotation_collection(self):
        self.details_tab.update_annotation_collection()
        self.geometry_tab.update_annotation()
        self.label_tab.update_annotation()
        self._set_active_shapes()

    def cancel(self):
        self.details_tab.cancel()
        self.geometry_tab.cancel()
        self.label_tab.cancel()

    def save(self):
        self.details_tab.save()
        self.geometry_tab.save()
        self.label_tab.save()


class LabelPanel(AnnotationPanel):
    def __init__(self, master, app_variables, **kwargs):
        """

        Parameters
        ----------
        master
            tkinter.Tk|tkinter.TopLevel
        app_variables : AppVariables
        kwargs
        """

        self._app_variables = app_variables  # type: AppVariables
        Frame.__init__(self, master, **kwargs)
        self.pack(expand=tkinter.TRUE, fill=tkinter.BOTH)

        self.name_panel = NamePanel(self, app_variables)  # type: NamePanel
        self.name_panel.grid(row=0, column=0, sticky='NSEW')

        self.tab_panel = LabelTabControl(self, app_variables)  # type: LabelTabControl
        self.tab_panel.grid(row=1, column=0, sticky='NSEW')

        self.button_panel = AnnotateButtons(self)  # type: AnnotateButtons
        self.button_panel.grid(row=2, column=0, sticky='NSEW')

        self.grid_rowconfigure(1, weight=1)
        self.grid_columnconfigure(0, weight=1)


class LabelCollectionViewer(AnnotationCollectionViewer):

    def __init__(self, master, app_variables, **kwargs):
        """

        Parameters
        ----------
        master
            The tkinter master element
        app_variables : AppVariables
        kwargs
            Optional keywords for the treeview initialization
        """

        self._app_variables = app_variables  # type: AppVariables
        self._directory_structure = None
        self._element_association = None

        if 'selectmode' not in kwargs:
            kwargs['selectmode'] = tkinter.BROWSE
        kwargs['columns'] = ('label', )
        TreeviewWithScrolling.__init__(self, master, **kwargs)
        self.heading('#0', text='Name')
        self.heading('#1', text='Label')
        self.update_annotation_collection()

    @property
    def annotation_collection(self):
        """
        FileLabelCollection : The file annotation collection
        """

        return self._app_variables.file_annotation_collection

    def _render_directory(self, directory, at_index='end'):
        if directory != '' and not self.exists(directory):
            stem, leaf = os.path.split(directory)
            self.insert(stem, at_index, directory, text=leaf, values=('', ))
        direct_info = self._directory_structure[directory]
        # render all the child directories
        for child_dir in direct_info['directories']:
            self._render_directory(child_dir)
        # render all the annotations
        for annotation_id in direct_info['elements']:
            self._render_annotation(annotation_id)

    def _render_annotation(self, the_id, at_index='end'):
        annotation = self.annotation_collection.annotations[the_id]
        parent = self._element_association[the_id]
        name = annotation.get_name()
        label_id = annotation.get_label_id()
        label = self._app_variables.get_label(label_id)
        if label is None:
            label = ''
        if name == annotation.uid and label_id != '':
            name = label
        self.insert(parent, at_index, the_id, text=name, values=(label, ))


class LabelCollectionPanel(AnnotationCollectionPanel):
    def __init__(self, master, app_variables, **kwargs):
        """

        Parameters
        ----------
        master
            The master widget
        app_variables : AppVariables
        kwargs
            keyword arguments passed through to the Frame constructor
        """

        Frame.__init__(self, master, **kwargs)

        self.button_panel = Frame(self, relief=tkinter.RIDGE, borderwidth=2)  # type: Frame
        self.new_button = Button(self.button_panel, text='New Annotation', width=28)  # type: Button
        self.new_button.grid(row=0, column=0, sticky='NW')
        self.edit_button = Button(self.button_panel, text='Edit Selected Annotation', width=28)  # type: Button
        self.edit_button.grid(row=1, column=0, sticky='NW')
        self.zoom_button = Button(self.button_panel, text='Zoom to Selected Annotation', width=28)  # type: Button
        self.zoom_button.grid(row=2, column=0, sticky='NW')
        self.move_button = Button(self.button_panel, text='Move Selected Annotation', width=28)  # type: Button
        self.move_button.grid(row=3, column=0, sticky='NW')
        self.button_panel.grid(row=0, column=0, sticky='NSEW')

        self.viewer = LabelCollectionViewer(self, app_variables)
        self.viewer.frame.grid(row=1, column=0, sticky='NSEW')
        self.grid_rowconfigure(1, weight=1)
        self.grid_columnconfigure(0, weight=1)


#########################
# the main tool

class LabelingTool(AnnotationTool):
    _NEW_ANNOTATION_TYPE = LabelFeature
    _NEW_FILE_ANNOTATION_TYPE = FileLabelCollection

    def __init__(self, master, reader=None, annotation_collection=None, **kwargs):
        """

        Parameters
        ----------
        master
            tkinter.Tk|tkinter.TopLevel
        reader : None|str|BaseReader|GeneralCanvasImageReader
        annotation_collection : None|str|FileLabelCollection
        kwargs
        """

        self.variables = AppVariables()  # type: AppVariables

        if 'sashrelief' not in kwargs:
            kwargs['sashrelief'] = tkinter.RIDGE
        if 'orient' not in kwargs:
            kwargs['orient'] = tkinter.HORIZONTAL
        PanedWindow.__init__(self, master, **kwargs)
        WidgetWithMetadata.__init__(self, master)
        self.pack(expand=tkinter.TRUE, fill=tkinter.BOTH)

        self.image_panel = ImagePanel(self)
        self.image_panel.canvas.set_canvas_size(400, 500)
        self.add(
            self.image_panel, width=400, height=700, padx=3, pady=3, sticky='NSEW', stretch=tkinter.FIRST)

        self.collection_panel = LabelCollectionPanel(self, self.variables)  # type: LabelCollectionPanel

        self.add(self.collection_panel, width=200, height=700, padx=3, pady=3, sticky='NSEW')

        # file menu
        self.menu_bar = tkinter.Menu()
        self.file_menu = tkinter.Menu(self.menu_bar, tearoff=0)
        self.file_menu.add_command(label="Open Image", command=self.select_image_file)
        self.file_menu.add_command(label="Open Directory", command=self.select_image_directory)
        self.file_menu.add_separator()
        self.file_menu.add_command(label="Open Existing Annotation File", command=self.select_annotation_file)
        self.file_menu.add_command(label="Create New Annotation File", command=self.create_new_annotation_file)
        self.file_menu.add_separator()
        self.file_menu.add_command(label="Save Annotation File", command=self.save_annotation_file)
        self.file_menu.add_separator()
        self.file_menu.add_command(label="Exit", command=self.exit)
        # edit popup menu
        self.edit_menu = tkinter.Menu(self.menu_bar, tearoff=0)
        self.edit_menu.add_command(label='Edit Label Schema', command=self.callback_edit_schema)
        self.edit_menu.add_separator()
        self.edit_menu.add_command(label='Replicate Selected', command=self.callback_replicate_feature)
        self.edit_menu.add_separator()
        self.edit_menu.add_command(label='Delete Selected', command=self.callback_delete_feature)

        # metadata popup menu
        self.metadata_menu = tkinter.Menu(self.menu_bar, tearoff=0)
        self.metadata_menu.add_command(label="Metaicon", command=self.metaicon_popup)
        self.metadata_menu.add_command(label="Metaviewer", command=self.metaviewer_popup)
        self._valid_data_shown = tkinter.IntVar(self, value=0)
        self.metadata_menu.add_checkbutton(
            label='ValidData', variable=self._valid_data_shown, command=self.show_valid_data)
        # configure menubar
        self.menu_bar.add_cascade(label="File", menu=self.file_menu)
        self.menu_bar.add_cascade(label="Edit", menu=self.edit_menu)
        self.menu_bar.add_cascade(label="Metadata", menu=self.metadata_menu)
        self.master.config(menu=self.menu_bar)

        # hide unwanted elements on the panel toolbars
        self.image_panel.hide_tools(['new_shape', 'select'])
        self.image_panel.hide_shapes()
        self.image_panel.hide_select_index()
        # disable tools until an image is selected
        self.image_panel.disable_tools()

        self.collection_panel.new_button.config(command=self.callback_new_annotation)
        self.collection_panel.edit_button.config(command=self.callback_popup_annotation)
        self.collection_panel.zoom_button.config(command=self.callback_zoom_to_feature)
        self.collection_panel.move_button.config(command=self.callback_move_feature)

        # set up context panel canvas event listeners
        self.image_panel.canvas.bind('<<ImageIndexPreChange>>', self.handle_image_index_prechange)
        self.image_panel.canvas.bind('<<ImageIndexChanged>>', self.handle_image_index_changed)
        self.image_panel.canvas.bind('<<ShapeCreate>>', self.shape_create_on_canvas)
        self.image_panel.canvas.bind('<<ShapeCoordsFinalized>>', self.shape_finalized_on_canvas)
        self.image_panel.canvas.bind('<<ShapeDelete>>', self.shape_delete_on_canvas)
        self.image_panel.canvas.bind('<<ShapeSelect>>', self.shape_selected_on_canvas)

        # set up the label_panel viewer event listeners
        self.collection_panel.viewer.bind('<<TreeviewSelect>>', self.feature_selected_on_viewer)

        self.annotate_popup = tkinter.Toplevel(master)
        self.annotate_popup.geometry('700x500')
        self.annotate = LabelPanel(self.annotate_popup, self.variables)  # type: LabelPanel
        self.annotate.hide_on_close()
        self.annotate_popup.withdraw()

        # bind actions/listeners from annotate popup
        self.annotate.tab_panel.geometry_tab.bind('<<GeometryPropertyChanged>>', self.geometry_selected_on_viewer)
        self.annotate.button_panel.delete_button.config(command=self.callback_delete_geometry)
        self.annotate.button_panel.cancel_button.config(command=self.callback_popup_cancel)
        self.annotate.button_panel.apply_button.config(command=self.callback_popup_apply)

        # bind the new geometry buttons explicitly
        self.annotate.tab_panel.geometry_tab.geometry_buttons.point.config(command=self.callback_new_point)
        self.annotate.tab_panel.geometry_tab.geometry_buttons.line.config(command=self.callback_new_line)
        self.annotate.tab_panel.geometry_tab.geometry_buttons.rectangle.config(command=self.callback_new_rect)
        self.annotate.tab_panel.geometry_tab.geometry_buttons.ellipse.config(command=self.callback_new_ellipse)
        self.annotate.tab_panel.geometry_tab.geometry_buttons.polygon.config(command=self.callback_new_polygon)

        # bind the label specifics submit button specifically
        self.annotate.tab_panel.label_tab.label_specifics.submit_button.config(command=self.callback_submit_label)

        self.set_reader(reader)
        self.set_annotations(annotation_collection)
        self.annotate.update_annotation_collection()

    def set_title(self):
        """
        Sets the window title.
        """

        file_name = None if self.variables.image_reader is None else self.variables.image_reader.file_name
        if file_name is None:
            the_title = "Labeling Tool"
        else:
            the_title = "Labeling Tool for {}".format(os.path.split(file_name)[1])
        self.winfo_toplevel().title(the_title)

    def select_annotation_file(self):
        if not self._verify_image_selected():
            return

        # prompt for any unsaved changes
        response = self._prompt_unsaved()
        if not response:
            return

        browse_dir, image_fname = os.path.split(self.image_file_name)
        # guess at a sensible initial file name
        init_file = '{}.labels.json'.format(image_fname)
        if not os.path.exists(os.path.join(browse_dir, init_file)):
            init_file = ''

        annotation_fname = askopenfilename(
            title='Select annotation file for image file {}'.format(image_fname),
            initialdir=browse_dir,
            initialfile=init_file,
            filetypes=[json_files, all_files])
        if annotation_fname in ['', ()]:
            return

        self.set_annotations(annotation_fname)

    def create_new_annotation_file(self):
        if not self._verify_image_selected():
            return

        # prompt for any unsaved changes
        response = self._prompt_unsaved()
        if not response:
            return

        label_schema_file = self._prompt_for_label_schema()
        if label_schema_file is None:
            label_schema = get_default_schema()
        else:
            label_schema = LabelSchema.from_file(label_schema_file)

        annotations = self._NEW_FILE_ANNOTATION_TYPE(label_schema)
        self.set_annotations(annotations)

    def _prompt_annotation_file_name(self):
        if self.image_file_name is not None:
            browse_dir, image_fname = os.path.split(self.image_file_name)
        else:
            browse_dir = self.variables.browse_directory
            image_fname = 'Unknown_Image'

        annotation_fname = asksaveasfilename(
            title='Select output annotation file name for image file {}'.format(image_fname),
            initialdir=browse_dir,
            initialfile='{}.labels.json'.format(image_fname),
            filetypes=[json_files, all_files])

        if annotation_fname in ['', ()]:
            annotation_fname = None
        return annotation_fname

    def _get_default_collection(self):
        return self._NEW_FILE_ANNOTATION_TYPE(label_schema=get_default_schema(), image_file_name=self.image_file_name)

    def _prompt_for_label_schema(self):
        browse_dir = self.variables.browse_directory
        label_schema_file = askopenfilename(
            title='Select Label Schema',
            initialdir=browse_dir,
            filetypes=[json_files, all_files])

        if label_schema_file in ['', ()]:
            label_schema_file = None
        return label_schema_file

    def callback_edit_schema(self):
        if self.variables.file_annotation_collection is None:
            return

        if self.variables.file_annotation_collection.annotations is not None and \
                len(self.variables.file_annotation_collection.annotations) > 0:
            showinfo(
                'label schema editing',
                message='You are opting to edit the label schema with existing labeled features.\n'
                        'Deleting any labels which are currently applied somewhere will result\n'
                        'in a non-viable label collection, and no error checking will be performed here.\n'
                        '\nDo not delete any labels from the schema unless you are sure.')

        root = tkinter.Toplevel(self.master)  # create a new toplevel with its own mainloop, so it's blocking
        # noinspection PyUnusedLocal
        tool = SchemaEditor(root, label_schema=self.variables.file_annotation_collection.label_schema)
        root.grab_set()
        root.wait_window()
        # update for any important state changes
        self.set_annotations(self.variables.file_annotation_collection)

    def callback_submit_label(self):
        """
        Submit the new label for the selected feature
        """

        if self.annotate.tab_panel.label_tab.label_specifics.callback_submit():
            self.annotate.update_annotation()
            self.collection_panel.update_annotation()


def main(reader=None, annotation=None):
    """
    Main method for initializing the labeling tool

    Parameters
    ----------
    reader : None|str|BaseReader|GeneralCanvasImageReader
    annotation : None|str|FileLabelCollection
    """

    root = tkinter.Tk()
    root.geometry("1000x800")

    the_style = ttk.Style()
    the_style.theme_use('classic')

    # noinspection PyUnusedLocal
    app = LabelingTool(root, reader=reader, annotation_collection=annotation)
    root.mainloop()


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser(
        description="Open the labeling tool with optional input file.",
        formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument(
        '-i', '--input', metavar='input', default=None,
        help='The path to the optional image file for opening.')
    parser.add_argument(
        '-a', '--annotation', metavar='annotation', default=None,
        help='The path to the optional annotation file. '
             'If the image input is not specified, then this has no effect. '
             'If both are specified, then a check will be performed that the '
             'annotation actually applies to the provided image.')
    this_args = parser.parse_args()

    main(reader=this_args.input, annotation=this_args.annotation)



"""
A general use generally abstract class for metaicon and metaviewer functionality.
"""

__author__ = "Thomas McCullough"
__classification__ = "UNCLASSIFIED"

import tkinter

from tk_builder.image_reader import CanvasImageReader

from sarpy_apps.supporting_classes.image_reader import ComplexCanvasImageReader, \
    DerivedCanvasImageReader
from sarpy_apps.supporting_classes.metaviewer import Metaviewer
from sarpy_apps.supporting_classes.metaicon.metaicon import MetaIcon


class WidgetWithMetadata(object):
    """
    Common use framework for handling metaicon and metaviewer functionality.
    """

    def __init__(self, master):
        """

        Parameters
        ----------
        master : tkinter.Tk|tkinter.Toplevel
        """

        self.metaicon_popup_panel = tkinter.Toplevel(master)
        self.metaicon = MetaIcon(self.metaicon_popup_panel)
        self.metaicon.hide_on_close()
        self.metaicon_popup_panel.withdraw()

        self.metaviewer_popup_panel = tkinter.Toplevel(master)
        self.metaviewer = Metaviewer(self.metaviewer_popup_panel)
        self.metaviewer.hide_on_close()
        self.metaviewer_popup_panel.withdraw()

    def _set_focus_on_metaicon_popup(self):
        self.metaicon_popup_panel.focus_set()
        self.metaicon_popup_panel.lift()

    def metaicon_popup(self):
        self.metaicon_popup_panel.deiconify()
        self._set_focus_on_metaicon_popup()

    def _set_focus_on_metaviewer_popup(self):
        self.metaviewer_popup_panel.focus_set()
        self.metaviewer_popup_panel.lift()

    def metaviewer_popup(self):
        self.metaviewer_popup_panel.deiconify()
        self._set_focus_on_metaviewer_popup()

    def populate_metaicon(self, image_reader):
        """
        Populate the metaicon.

        Parameters
        ----------
        image_reader : CanvasImageReader
        """

        if image_reader is None:
            self.metaicon.make_empty()

        if isinstance(image_reader, ComplexCanvasImageReader):
            self.metaicon.create_from_reader(image_reader.base_reader, index=image_reader.index)
        elif isinstance(image_reader, DerivedCanvasImageReader):
            self.metaicon.create_from_reader(image_reader.base_reader, index=image_reader.index)
        else:
            self.metaicon.make_empty()

    def populate_metaviewer(self, image_reader):
        """
        Populate the metaviewer.

        Parameters
        ----------
        image_reader : CanvasImageReader
        """

        if image_reader is None:
            self.metaviewer.empty_entries()

        if isinstance(image_reader, (ComplexCanvasImageReader, DerivedCanvasImageReader)):
            self.metaviewer.populate_from_reader(image_reader.base_reader)
        else:
            self.metaviewer.empty_entries()



# -*- coding: utf-8 -*-
"""
This module provides a full image frequency support visualization tool, intended
primarily for use in the validation tool.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

import logging
from typing import Optional
from tempfile import mkstemp
import os

import tkinter
from tkinter import ttk
from tkinter.filedialog import askopenfilenames, askdirectory
from tkinter.messagebox import showinfo

import numpy
from matplotlib import pyplot
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, \
    NavigationToolbar2Tk

from tk_builder.base_elements import TypedDescriptor, StringDescriptor
from tk_builder.image_reader import NumpyCanvasImageReader
from tk_builder.panels.image_panel import ImagePanel
from tk_builder.widgets.basic_widgets import Frame

from sarpy_apps.supporting_classes.file_filters import common_use_collection
from sarpy_apps.supporting_classes.image_reader import ComplexCanvasImageReader, SICDTypeCanvasImageReader
from sarpy_apps.supporting_classes.widget_with_metadata import WidgetWithMetadata

from sarpy.io.complex.base import FlatSICDReader
from sarpy.processing.fft_base import fft_sicd, fft2_sicd, fftshift
from sarpy.processing.normalize_sicd import DeskewCalculator
from sarpy.io.complex.base import SICDTypeReader

logger = logging.getLogger(__name__)


def create_deskewed_transform(reader, dimension=0, suffix='.sarpy.cache'):
    """
    Performs the Fourier transform of the deskewed entirety of the given
    ComplexImageReader contents.

    Parameters
    ----------
    reader : SICDTypeCanvasImageReader
        The reader object.
    dimension : int
        One of [0, 1], which dimension to deskew along.
    suffix : None|str
        The suffix for the created file name (created using the tempfile module).

    Returns
    -------
    (str, numpy.ndarray, numpy.ndarray)
        A file name, numpy memmap of the given object, and mean along the given dimension.
        Care should be taken to ensure that the file is deleted when the usage is complete.
    """

    # set up a true file for the memmap
    # NB: it should be noted that the tempfile usage which clean themselves up
    #     cannot (as of 2021-04-23) be opened multiple times on Windows, which
    #     means that such a "file" cannot be used in conjunction with a numpy
    #     memmap.
    data_size = reader.data_size
    sicd = reader.get_sicd()
    _, file_name = mkstemp(suffix=suffix, text=False)
    logger.debug('Creating temp file % s' % file_name)
    # set up the memmap
    memmap = numpy.memmap(file_name, dtype='complex64', mode='r+', offset=0, shape=data_size)
    calculator = DeskewCalculator(
        reader.base_reader, dimension=dimension, index=reader.index,
        apply_deskew=True, apply_deweighting=False, apply_off_axis=False)
    mean_value = numpy.zeros((data_size[0], ), dtype='float64') if dimension == 0 else \
        numpy.zeros((data_size[1],), dtype='float64')

    # we'll proceed in blocks of approximately this number of pixels
    pixels_threshold = 2**20
    # is our whole reader sufficiently small to just do it all in one fell-swoop?
    if data_size[0]*data_size[1] <= 4*pixels_threshold:
        data = fftshift(fft2_sicd(calculator[:, :], sicd))
        memmap[:, :] = data
        mean_value[:] = numpy.mean(numpy.abs(data), axis=1-dimension)
        return file_name, memmap, mean_value

    # fetch full rows, and transform then shift along the row direction
    block_size = int(numpy.ceil(pixels_threshold/data_size[1]))
    start_col = 0
    while start_col < data_size[1]:
        end_col = min(start_col+block_size, data_size[1])
        data = fftshift(fft_sicd(calculator[:, start_col:end_col], 0, sicd), axes=0)
        memmap[:, start_col:end_col] = data
        if dimension == 0:
            mean_value += numpy.sum(numpy.abs(data), axis=1)
        start_col = end_col
    # fetch full columns, and transform then shift along the column direction
    block_size = int(numpy.ceil(pixels_threshold/data_size[0]))
    start_row = 0
    while start_row < data_size[0]:
        end_row = min(start_row+block_size, data_size[0])
        data = fftshift(fft_sicd(memmap[start_row:end_row, :], 1, sicd), axes=1)
        memmap[start_row:end_row, :] = data
        if dimension == 1:
            mean_value += numpy.sum(numpy.abs(data), axis=0)
        start_row = end_row

    if dimension == 0:
        mean_value /= data_size[1]
    else:
        mean_value /= data_size[0]
    return file_name, memmap, mean_value


class AppVariables(object):
    browse_directory = StringDescriptor(
        'browse_directory', default_value=os.path.expanduser('~'),
        docstring='The directory for browsing for file selection.')  # type: str
    remap_type = StringDescriptor(
        'remap_type', default_value='', docstring='')  # type: str
    image_reader = TypedDescriptor(
        'image_reader', SICDTypeCanvasImageReader, docstring='')  # type: SICDTypeCanvasImageReader

    row_fourier_reader = TypedDescriptor(
        'row_fourier_reader', ComplexCanvasImageReader,
        docstring='The row deskewed fourier transformed reader')  # type: ComplexCanvasImageReader
    row_fourier_file = StringDescriptor(
        'row_fourier_file',
        docstring='The row deskewed fourier transformed reader file')  # type: Optional[str]
    # NB: we are saving this state in order to properly clean up
    column_fourier_reader = TypedDescriptor(
        'column_fourier_reader', ComplexCanvasImageReader,
        docstring='The column deskewed fourier transformed reader')  # type: ComplexCanvasImageReader
    column_fourier_file = StringDescriptor(
        'row_fourier_file',
        docstring='The column deskewed fourier transformed reader file')  # type: Optional[str]
    # NB: we are saving this state in order to properly clean up

    derived_row_weights = None  # the derived weights for the row
    scaled_row_mean = None  # the scaled mean Fourier transform of the row deskewed data
    derived_column_weights = None  # the derived weights for the column
    scaled_column_mean = None  # the scaled mean Fourier transform of the column deskewed data

    def __del__(self):
        # clean up files, because we might need to do so
        if self.row_fourier_file is not None \
                and os.path.exists(self.row_fourier_file):
            os.remove(self.row_fourier_file)
            logger.debug('(variables) Removing temp file % s' % self.row_fourier_file)
            self.row_fourier_file = None

        if self.column_fourier_file is not None and \
                os.path.exists(self.column_fourier_file):
            os.remove(self.column_fourier_file)
            logger.debug('(variables)  Removing temp file % s' % self.column_fourier_file)
            self.column_fourier_file = None


class FullFrequencySupportTool(Frame, WidgetWithMetadata):
    def __init__(self, primary, reader=None, **kwargs):
        """

        Parameters
        ----------
        primary
            tkinter.Tk|tkinter.TopLevel
        reader : None|str|SICDTypeReader|SICDTypeCanvasImageReader
        kwargs
        """

        self.root = primary
        self.variables = AppVariables()

        Frame.__init__(self, primary, **kwargs)
        WidgetWithMetadata.__init__(self, primary)

        self.row_centered_image_panel = ImagePanel(self, borderwidth=2, relief=tkinter.RIDGE)  # type: ImagePanel
        self.row_centered_image_panel.grid(row=0, column=0, sticky='NSEW')

        self.column_centered_image_panel = ImagePanel(self, borderwidth=2, relief=tkinter.RIDGE)  # type: ImagePanel
        self.column_centered_image_panel.grid(row=0, column=1, sticky='NSEW')
        self.grid_rowconfigure(0, weight=1)
        self.grid_columnconfigure(0, weight=1)
        self.grid_columnconfigure(1, weight=1)

        self.set_title()

        # define menus
        menubar = tkinter.Menu()
        # file menu
        filemenu = tkinter.Menu(menubar, tearoff=0)
        filemenu.add_command(label="Open Image", command=self.callback_select_files)
        filemenu.add_command(label="Open Directory", command=self.callback_select_directory)
        filemenu.add_separator()
        filemenu.add_command(label="Exit", command=self.exit)
        # menus for informational popups
        popups_menu = tkinter.Menu(menubar, tearoff=0)
        popups_menu.add_command(label="Weight Plots", command=self.create_weights_plot)
        popups_menu.add_command(label="Metaicon", command=self.metaicon_popup)
        popups_menu.add_command(label="Metaviewer", command=self.metaviewer_popup)
        # ensure menus cascade
        menubar.add_cascade(label="File", menu=filemenu)
        menubar.add_cascade(label="Metadata", menu=popups_menu)

        # handle packing
        self.pack(fill=tkinter.BOTH, expand=tkinter.YES)
        primary.config(menu=menubar)

        # hide extraneous tool elements
        self.row_centered_image_panel.hide_tools(['shape_drawing', 'select', 'coords', 'measure'])
        self.row_centered_image_panel.hide_shapes()
        self.row_centered_image_panel.hide_select_index()

        self.column_centered_image_panel.hide_tools(['shape_drawing', 'select', 'coords', 'measure'])
        self.column_centered_image_panel.hide_shapes()
        self.column_centered_image_panel.hide_select_index()

        self.update_reader(reader)
        # TODO: allow changing the image index somewhere?
        #   bind the handle_image_index_changed method

    def set_title(self):
        """
        Sets the window title.
        """

        file_name = None if self.variables.image_reader is None else self.variables.image_reader.file_name
        if file_name is None:
            the_title = "Full Frequency Support Tool"
        elif isinstance(file_name, (list, tuple)):
            the_title = "Full Frequency Support Tool, Multiple Files"
        else:
            the_title = "Full Frequency Support Tool for {}".format(os.path.split(file_name)[1])
        self.winfo_toplevel().title(the_title)

    def exit(self):
        self._delete_files()
        self.root.destroy()

    def _delete_files(self):
        """
        This is a helper function for cleaning up our state.
        """

        if not hasattr(self, 'variables') or self.variables is None:
            return

        self.variables.row_fourier_reader = None
        self.variables.column_fourier_reader = None
        if self.variables.row_fourier_file is not None \
                and os.path.exists(self.variables.row_fourier_file):
            os.remove(self.variables.row_fourier_file)
            logger.debug('Removing temp file % s' % self.variables.row_fourier_file)

        if self.variables.column_fourier_file is not None and \
                os.path.exists(self.variables.column_fourier_file):
            os.remove(self.variables.column_fourier_file)
            logger.debug('Removing temp file % s' % self.variables.column_fourier_file)

        self.variables.row_fourier_file = None
        self.variables.column_fourier_file = None

    def _clear_display(self):
        """
        Clear the row and column Fourier transform displays
        """

        self._delete_files()
        junk_data = numpy.zeros((100, 100), dtype='uint8')
        self.row_centered_image_panel.set_image_reader(NumpyCanvasImageReader(junk_data))
        self.column_centered_image_panel.set_image_reader(NumpyCanvasImageReader(junk_data))

    def _calculate_fourier_data(self):
        def set_row_data():
            # calculate the fourier transform with deskew in the row direction
            row_file, row_memmap, row_mean_value = create_deskewed_transform(self.variables.image_reader, dimension=0)
            self.variables.row_fourier_file = row_file
            self.variables.row_fourier_reader = ComplexCanvasImageReader(
                FlatSICDReader(self.variables.image_reader.get_sicd(), row_memmap))
            self.row_centered_image_panel.set_image_reader(self.variables.row_fourier_reader)

            draw_deltak_lines(self.row_centered_image_panel.canvas, 'row')

            # rescale the row_mean_value so that the smoothed max value is essentially 1
            if row_mean_value.size < 200:
                the_max = numpy.amax(row_mean_value)
            else:
                the_size = int(numpy.ceil(row_mean_value.size/200.))
                smoothed = numpy.convolve(
                    row_mean_value, numpy.full((the_size, ), 1./the_size, dtype='float64'),
                    mode='valid')
                the_max = numpy.amax(smoothed)
            row_mean_value /= the_max
            self.variables.scaled_row_mean = row_mean_value
            # construct the proper weights and prepare information for weight plotting
            self.variables.derived_row_weights = the_sicd.Grid.Row.define_weight_function(populate=False)

        def set_col_data():
            # calculate the fourier transform with deskew in the column direction
            col_file, col_memmap, col_mean_value = create_deskewed_transform(self.variables.image_reader, dimension=1)
            self.variables.column_fourier_file = col_file
            self.variables.column_fourier_reader = ComplexCanvasImageReader(
                FlatSICDReader(self.variables.image_reader.get_sicd(), col_memmap))
            self.column_centered_image_panel.set_image_reader(self.variables.column_fourier_reader)

            draw_deltak_lines(self.column_centered_image_panel.canvas, 'column')

            # rescale the row_mean_value so that the smoothed max value is essentially 1
            if col_mean_value.size < 200:
                the_max = numpy.amax(col_mean_value)
            else:
                the_size = int(numpy.ceil(col_mean_value.size / 200.))
                smoothed = numpy.convolve(col_mean_value, numpy.full((the_size,), 1./the_size, dtype='float64'),
                                          mode='valid')
                the_max = numpy.amax(smoothed)
            col_mean_value /= the_max
            self.variables.scaled_column_mean = col_mean_value
            # construct the proper weights and prepare information for weight plotting
            self.variables.derived_column_weights = the_sicd.Grid.Col.define_weight_function(populate=False)

        def draw_deltak_lines(canvas, dimension):
            if dimension == 'row':
                # populate row as full bandwidth
                row_deltak1 = (row_count - 1)*(0.5 - 0.5*the_sicd.Grid.Row.SS*the_sicd.Grid.Row.ImpRespBW) + 1
                row_deltak2 = (row_count - 1)*(0.5 + 0.5*the_sicd.Grid.Row.SS*the_sicd.Grid.Row.ImpRespBW) + 1
                # calculate the column deltak1/deltak2 values
                col_deltak1 = (col_count - 1) * (0.5 + the_sicd.Grid.Col.SS*the_sicd.Grid.Col.DeltaK1) + 1
                col_deltak2 = (col_count - 1) * (0.5 + the_sicd.Grid.Col.SS*the_sicd.Grid.Col.DeltaK2) + 1
            elif dimension == 'column':
                # calculate the row deltak1/deltak2 values
                row_deltak1 = (row_count - 1)*(0.5 + the_sicd.Grid.Row.SS*the_sicd.Grid.Row.DeltaK1) + 1
                row_deltak2 = (row_count - 1)*(0.5 + the_sicd.Grid.Row.SS*the_sicd.Grid.Row.DeltaK2) + 1
                # populate column as full bandwidth
                col_deltak1 = (col_count - 1) * (0.5 - 0.5*the_sicd.Grid.Col.SS*the_sicd.Grid.Col.ImpRespBW) + 1
                col_deltak2 = (col_count - 1) * (0.5 + 0.5*the_sicd.Grid.Col.SS*the_sicd.Grid.Col.ImpRespBW) + 1
            else:
                raise ValueError('Unrecognized dimension argument `{}`'.format(dimension))

            # draw the row deltak1/deltak2 lines
            row_deltak1_id = canvas.create_new_line(
                (0, 0, 0, 0), make_current=False, increment_color=False, color='red')
            canvas.modify_existing_shape_using_image_coords(
                row_deltak1_id, (row_deltak1, 0, row_deltak1, col_count))
            deltak2_id = canvas.create_new_line(
                (0, 0, 0, 0), make_current=False, increment_color=False, color='red')
            canvas.modify_existing_shape_using_image_coords(
                deltak2_id, (row_deltak2, 0, row_deltak2, col_count))

            # draw the column deltak1/deltak2 lines
            col_deltak1_id = canvas.create_new_line(
                (0, 0, 0, 0), make_current=False, increment_color=False, color='red')
            canvas.modify_existing_shape_using_image_coords(
                col_deltak1_id, (0, col_deltak1, row_count, col_deltak1))
            deltak2_id = canvas.create_new_line(
                (0, 0, 0, 0), make_current=False, increment_color=False, color='red')
            canvas.modify_existing_shape_using_image_coords(
                deltak2_id, (0, col_deltak2, row_count, col_deltak2))

        # delete any previous state variables and clear displays
        self._clear_display()
        self.variables.scaled_row_mean = None
        self.variables.derived_row_weights = None
        self.variables.scaled_column_mean = None
        self.variables.derived_column_weights = None
        if self.variables.image_reader is None:
            return

        self.update_idletasks()
        the_sicd = self.variables.image_reader.get_sicd()
        row_count = the_sicd.ImageData.NumRows
        col_count = the_sicd.ImageData.NumCols

        set_row_data()
        self.update_idletasks()
        set_col_data()

    def create_weights_plot(self):
        """
        Create a matplotlib (using the user default backend) of the weight
        information.
        """

        if self.variables.image_reader is None or \
                self.variables.scaled_row_mean is None or \
                self.variables.scaled_column_mean is None:
            return  # nothing to be done currently

        the_sicd = self.variables.image_reader.get_sicd()
        fig, axs = pyplot.subplots(nrows=2, ncols=1)

        the_title = 'Weight information for file {}'.format(
            os.path.split(self.variables.image_reader.file_name)[1])
        fig.suptitle(the_title)

        # plot the row information
        axs[0].set_ylabel('Row Information')
        axs[0].set_xlabel('Krow (cycles/meter)')
        axs[0].plot(
            numpy.linspace(-0.5/the_sicd.Grid.Row.SS,
                           0.5/the_sicd.Grid.Row.SS,
                           self.variables.scaled_row_mean.size),
            self.variables.scaled_row_mean, 'b', lw=1, label='Observed Data')
        if self.variables.derived_row_weights is not None:
            axs[0].plot(numpy.linspace(-0.5*the_sicd.Grid.Row.ImpRespBW,
                                       0.5*the_sicd.Grid.Row.ImpRespBW,
                                       self.variables.derived_row_weights.size),
                        self.variables.derived_row_weights,
                        'g--', lw=3, label='Row Derived Weights')
        if the_sicd.Grid.Row.WgtFunct is not None:
            axs[0].plot(numpy.linspace(-0.5*the_sicd.Grid.Row.ImpRespBW,
                                       0.5*the_sicd.Grid.Row.ImpRespBW,
                                       the_sicd.Grid.Row.WgtFunct.size),
                        the_sicd.Grid.Row.WgtFunct,
                        'r:', lw=3, label='Row.WgtFunct')
        axs[0].set_xlim(min(-0.5/the_sicd.Grid.Row.SS, -0.5*the_sicd.Grid.Row.ImpRespBW),
                        max(0.5/the_sicd.Grid.Row.SS, 0.5*the_sicd.Grid.Row.ImpRespBW))
        axs[0].legend(loc='upper right')

        axs[1].set_ylabel('Column Information')
        axs[1].set_xlabel('Kcol (cycles/meter)')
        axs[1].plot(
            numpy.linspace(-0.5/the_sicd.Grid.Col.SS,
                           0.5/the_sicd.Grid.Col.SS,
                           self.variables.scaled_column_mean.size),
            self.variables.scaled_column_mean, 'b', lw=1, label='Observed Data')
        if self.variables.derived_column_weights is not None:
            axs[1].plot(numpy.linspace(-0.5*the_sicd.Grid.Col.ImpRespBW,
                                       0.5*the_sicd.Grid.Col.ImpRespBW,
                                       self.variables.derived_column_weights.size),
                        self.variables.derived_column_weights,
                        'g--', lw=3, label='Col Derived Weights')
        if the_sicd.Grid.Col.WgtFunct is not None:
            axs[1].plot(numpy.linspace(-0.5*the_sicd.Grid.Col.ImpRespBW,
                                       0.5*the_sicd.Grid.Col.ImpRespBW,
                                       the_sicd.Grid.Col.WgtFunct.size),
                        the_sicd.Grid.Col.WgtFunct,
                        'r:', lw=3, label='Col.WgtFunct')
        axs[1].set_xlim(min(-0.5/the_sicd.Grid.Col.SS, -0.5*the_sicd.Grid.Col.ImpRespBW),
                        max(0.5/the_sicd.Grid.Col.SS, 0.5*the_sicd.Grid.Col.ImpRespBW))
        axs[1].legend(loc='upper right')

        # create a toplevel, and put our figure inside it
        root = tkinter.Toplevel(self.root)

        root.wm_title("Embedding in Tk")

        canvas = FigureCanvasTkAgg(fig, master=root)  # A tk.DrawingArea.
        canvas.draw()
        canvas.get_tk_widget().pack(side=tkinter.TOP, fill=tkinter.BOTH, expand=1)

        toolbar = NavigationToolbar2Tk(canvas, root)
        toolbar.update()
        canvas.get_tk_widget().pack(side=tkinter.TOP, fill=tkinter.BOTH, expand=1)

        # grab the focus, so this is blocking
        root.grab_set()
        root.wait_window()

    def handle_image_index_changed(self):
        """
        Handle that the image index has changed.
        """

        self.my_populate_metaicon()
        self._calculate_fourier_data()

    def update_reader(self, the_reader, update_browse=None):
        """
        Update the reader.

        Parameters
        ----------
        the_reader : None|str|SICDTypeReader|SICDTypeCanvasImageReader
        update_browse : None|str
        """

        if the_reader is None:
            return

        if update_browse is not None:
            self.variables.browse_directory = update_browse
        elif isinstance(the_reader, str):
            self.variables.browse_directory = os.path.split(the_reader)[0]

        if isinstance(the_reader, str):
            the_reader = SICDTypeCanvasImageReader(the_reader)

        if isinstance(the_reader, SICDTypeReader):
            the_reader = SICDTypeCanvasImageReader(the_reader)

        if not isinstance(the_reader, SICDTypeCanvasImageReader):
            raise TypeError('Got unexpected input for the reader')

        # change the tool to view
        self.row_centered_image_panel.canvas.current_tool = 'VIEW'
        self.row_centered_image_panel.canvas.current_tool = 'VIEW'
        # update the reader
        self.variables.image_reader = the_reader
        self.variables.image_index = 0
        self.set_title()
        # refresh appropriate GUI elements
        self._calculate_fourier_data()
        self.my_populate_metaicon()
        self.my_populate_metaviewer()

    def callback_select_files(self):
        fnames = askopenfilenames(initialdir=self.variables.browse_directory, filetypes=common_use_collection)
        if fnames is None or fnames in ['', ()]:
            return

        if len(fnames) == 1:
            the_reader = SICDTypeCanvasImageReader(fnames[0])
        else:
            the_reader = SICDTypeCanvasImageReader(fnames)

        if the_reader is None:
            showinfo('Opener not found',
                     message='File {} was not successfully opened as a SICD type '
                             'file.'.format(fnames))
            return
        self.update_reader(the_reader, update_browse=os.path.split(fnames[0])[0])

    def callback_select_directory(self):
        dirname = askdirectory(initialdir=self.variables.browse_directory, mustexist=True)
        if dirname is None or dirname in [(), '']:
            return

        the_reader = SICDTypeCanvasImageReader(dirname)
        self.update_reader(the_reader, update_browse=os.path.split(dirname)[0])

    def my_populate_metaicon(self):
        """
        Populate the metaicon.
        """

        self.populate_metaicon(self.variables.image_reader)

    def my_populate_metaviewer(self):
        """
        Populate the metaviewer.
        """

        self.populate_metaviewer(self.variables.image_reader)

    def __del__(self):
        self._delete_files()


def main(reader=None):
    """
    Main method for initializing the tool

    Parameters
    ----------
    reader : None|str|SICDTypeReader|SICDTypeCanvasImageReader
    """

    root = tkinter.Tk()

    the_style = ttk.Style()
    the_style.theme_use('classic')

    app = FullFrequencySupportTool(root, reader=reader)
    root.geometry("1000x1000")
    root.mainloop()


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser(
        description="Open the full support frequency analysis tool with optional input file.",
        formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument(
        '-i', '--input', metavar='input', default=None, help='The path to the optional image file for opening.')
    args = parser.parse_args()

    main(reader=args.input)



# -*- coding: utf-8 -*-
"""
This module provides a tool for doing some basic validation for a SICD file.
"""

__classification__ = "UNCLASSIFIED"
__author__ = "Thomas McCullough"

import logging
import os

import tkinter
from tkinter import ttk
from tkinter.filedialog import askdirectory, askopenfilename, asksaveasfilename
from tkinter.messagebox import showinfo
from tkinter.scrolledtext import ScrolledText

from tk_builder.base_elements import StringDescriptor, TypedDescriptor
from tk_builder.file_filters import create_filter_entry, all_files
from tk_builder.logger import TextHandler
from tk_builder.panel_builder import WidgetPanel, WidgetPanelNoLabel
from tk_builder.widgets.basic_widgets import Label, CheckButton, Text, Button
from tk_builder.widgets.widget_descriptors import LabelDescriptor, CheckButtonDescriptor, \
    TextDescriptor, ButtonDescriptor

from sarpy_apps.apps.aperture_tool import RegionSelection
from sarpy_apps.apps.local_support_tool import LocalFrequencySupportTool
from sarpy_apps.apps.full_support_tool import FullFrequencySupportTool
from sarpy_apps.apps.rcs_tool import RCSTool
from sarpy_apps.supporting_classes.file_filters import common_use_collection
from sarpy_apps.supporting_classes.image_reader import SICDTypeCanvasImageReader
from sarpy_apps.supporting_classes.widget_with_metadata import WidgetWithMetadata

from sarpy.io.product.kmz_product_creation import create_kmz_view
from sarpy.io.complex.base import SICDTypeReader
from sarpy.io.complex.sicd import SICDReader
from sarpy.consistency.sicd_consistency import check_file
from sarpy.io.general.base import SarpyIOError


class _Feedback(WidgetPanel):
    _widget_list = (
        ('title_label', ),
        ('acceptable_label', 'acceptable_button'),
        ('feedback_label', 'feedback_text'),
        ('cancel_button', 'submit_button'))

    title_label = LabelDescriptor(
        'title_label', default_text='',
        docstring='The overall title')  # type: Label
    acceptable_label = LabelDescriptor(
        'acceptable_label', default_text='acceptable?', docstring='')  # type: Label
    acceptable_button = CheckButtonDescriptor(
        'acceptable_button', docstring='')  # type: CheckButton
    feedback_label = LabelDescriptor(
        'feedback_label', default_text='feedback', docstring='')  # type: Label
    feedback_text = TextDescriptor(
        'feedback_text', docstring='The widget to provide log information')  # type: Text
    cancel_button = ButtonDescriptor(
        'cancel_button', default_text='Cancel', docstring='')  # type: Button
    submit_button = ButtonDescriptor(
        'submit_button', default_text='Submit', docstring='')  # type: Button

    def __init__(self, root, title_text):
        """

        Parameters
        ----------
        root : tkinter.Toplevel
        title_text : str
        """

        WidgetPanel.__init__(self, root)
        self.init_w_rows()
        self.title_label.set_text(title_text)


class FeedbackPopup(object):
    """
    Class enabling feedback from validation steps
    """

    def __init__(self, primary, title_text):
        self.use_feedback = True  # type: bool
        self.acceptable = False  # type: bool
        self.feedback_text = ''  # type: str

        self.root = tkinter.Toplevel(primary)
        self.widget = _Feedback(self.root, title_text)
        self.widget.set_text('Feedback')

        # label appearance
        self.widget.title_label.config(anchor=tkinter.CENTER, relief=tkinter.RIDGE)
        self.widget.acceptable_label.config(anchor=tkinter.CENTER, relief=tkinter.RIDGE)
        self.widget.feedback_label.config(anchor=tkinter.CENTER, relief=tkinter.RIDGE)

        # set up callbacks
        self.widget.cancel_button.config(command=self.callback_cancel)
        self.widget.submit_button.config(command=self.callback_submit)

    def callback_cancel(self):
        self.use_feedback = False
        self.root.destroy()

    def callback_submit(self):
        self.use_feedback = True
        self.acceptable = self.widget.acceptable_button.instate(['selected'])
        self.feedback_text = self.widget.feedback_text.get('1.0', 'end-1c')
        self.root.destroy()


class _Buttons(WidgetPanelNoLabel):
    """
    The panel of buttons for validation tasks.
    """

    _widget_list = (
        ('local_fs_label', 'local_fs_button'),
        ('full_fs_label', 'full_fs_button'),
        ('sign_label', 'sign_button'),
        ('noise_label', 'noise_button'),
        ('geolocation_label', 'geolocation_button'))

    local_fs_label = LabelDescriptor(
        'local_fs_label', default_text='Perform local frequency support analysis',
        docstring='')  # type: Label
    local_fs_button = ButtonDescriptor(
        'local_fs_button', default_text='Frequency Support Tool',
        docstring='')  # type: Button

    full_fs_label = LabelDescriptor(
        'full_fs_label', default_text='Perform full image frequency analysis',
        docstring='')  # type: Label
    full_fs_button = ButtonDescriptor(
        'full_fs_button', default_text='Frequency Analysis Tool',
        docstring='')  # type: Button

    sign_label = LabelDescriptor(
        'sign_label', default_text='Perform Fourier analysis',
        docstring='')  # type: Label
    sign_button = ButtonDescriptor(
        'sign_button', default_text='Aperture Tool',
        docstring='')  # type: Button

    noise_label = LabelDescriptor(
        'noise_label', default_text='Perform noise analysis',
        docstring='')  # type: Label
    noise_button = ButtonDescriptor(
        'noise_button', default_text='RCS Tool',
        docstring='')  # type: Button

    geolocation_label = LabelDescriptor(
        'geolocation_label', default_text='Perform basic geolocation analysis',
        docstring='')  # type: Label
    geolocation_button = ButtonDescriptor(
        'geolocation_button', default_text='Create kmz',
        docstring='')  # type: Button

    def __init__(self, parent):
        """

        Parameters
        ----------
        parent
        """

        WidgetPanelNoLabel.__init__(self, parent)
        self.init_w_rows()


class AppVariables(object):
    browse_directory = StringDescriptor(
        'browse_directory', default_value=os.path.expanduser('~'),
        docstring='The directory for browsing for file selection.')  # type: str
    image_reader = TypedDescriptor(
        'image_reader', SICDTypeCanvasImageReader, docstring='')  # type: SICDTypeCanvasImageReader


class ValidationTool(tkinter.PanedWindow, WidgetWithMetadata):
    _widget_list = ("button_panel", "text_log_widget")
    button_panel = TypedDescriptor(
        'button_panel', _Buttons, docstring='the button panel')  # type: _Buttons
    text_log_widget = TypedDescriptor(
        'text_log_widget', ScrolledText, docstring='the log display')  # type: ScrolledText

    def __init__(self, primary, reader=None, **kwargs):
        """

        Parameters
        ----------
        primary : tkinter.Toplevel|tkinter.Tk
        reader : None|str|SICDTypeReader|SICDTypeCanvasImageReader
        """

        self.variables = AppVariables()

        if 'sashrelief' not in kwargs:
            kwargs['sashrelief'] = tkinter.RIDGE
        if 'orient' not in kwargs:
            kwargs['orient'] = tkinter.VERTICAL

        tkinter.PanedWindow.__init__(self, primary, **kwargs)
        WidgetWithMetadata.__init__(self, primary)
        self.pack(expand=tkinter.TRUE, fill=tkinter.BOTH)

        # handle packing manually
        self.button_panel = _Buttons(self)
        self.add(self.button_panel, width=700, height=300, padx=5, pady=5, sticky=tkinter.NSEW)

        # create the scrolled text widget for logging output
        self.text_log_widget = ScrolledText(self)  # TODO: other configuration?
        self.add(self.text_log_widget, width=700, height=400, padx=5, pady=5, sticky=tkinter.NSEW)

        # set the logging handler for the validation logger to log to our widget
        self.log_handler = TextHandler(self.text_log_widget)  # type: TextHandler
        self.log_handler.setFormatter(logging.Formatter('%(levelname)s:%(asctime)s - %(message)s', '%Y-%m-%dT%H:%M:%S'))
        # attach this handler to the validation logger
        self.logger = logging.getLogger('validation')
        self.logger.setLevel('INFO')
        self.logger.addHandler(self.log_handler)

        self.set_title()

        # define menus
        menubar = tkinter.Menu()
        # file menu
        filemenu = tkinter.Menu(menubar, tearoff=0)
        filemenu.add_command(label="Open Image", command=self.callback_select_file)
        filemenu.add_command(label="Open Directory", command=self.callback_select_directory)
        filemenu.add_separator()
        filemenu.add_command(label="Save Log", command=self.save_log)
        filemenu.add_command(label="Exit", command=self.exit)
        # menus for informational popups
        popups_menu = tkinter.Menu(menubar, tearoff=0)
        popups_menu.add_command(label="Metaicon", command=self.metaicon_popup)
        popups_menu.add_command(label="Metaviewer", command=self.metaviewer_popup)
        # ensure menus cascade
        menubar.add_cascade(label="File", menu=filemenu)
        menubar.add_cascade(label="Metadata", menu=popups_menu)

        self.master.config(menu=menubar)

        # set the callbacks for the button panel
        self.button_panel.local_fs_button.config(command=self.callback_local_fs)
        self.button_panel.full_fs_button.config(command=self.callback_full_fs)
        self.button_panel.sign_button.config(command=self.callback_sign)
        self.button_panel.noise_button.config(command=self.callback_noise)
        self.button_panel.geolocation_button.config(command=self.callback_geolocation)

        self.update_reader(reader)

    def _verify_reader(self):
        # type: () -> bool
        if self.variables.image_reader is not None:
            return True

        showinfo('No complex image selected', message='First, select a complex image for this functionality.')
        return False

    def _get_and_log_feedback(self, title_text):
        # type: (str) -> None
        feedback = FeedbackPopup(self.master, title_text)
        feedback.root.grab_set()
        feedback.root.wait_window()
        if not feedback.use_feedback:
            return

        feedback_text = feedback.feedback_text.strip()
        if feedback_text != '':
            feedback_text += '\n'
        if feedback.acceptable:
            self.logger.info('{} acceptable\n{}'.format(title_text, feedback_text))
        else:
            self.logger.error('{} unacceptable\n{}'.format(title_text, feedback_text))

    def set_title(self):
        """
        Sets the window title.
        """

        file_name = None if self.variables.image_reader is None else self.variables.image_reader.file_name
        if file_name is None:
            the_title = "Validation Tool"
        else:
            the_title = "Image Viewer for {}".format(os.path.split(file_name)[1])
        self.winfo_toplevel().title(the_title)

    def exit(self):
        self.master.destroy()

    def save_log(self):
        if self.variables.image_reader is None or self.log_handler is None:
            return

        save_fname = asksaveasfilename(
            initialdir=self.variables.browse_directory,
            title="Save validation log output to location?",
            initialfile='{}.validation.txt'.format(os.path.splitext(self.variables.image_reader.file_name)[0]),
            filetypes=[create_filter_entry('Log files', '.log .txt'), all_files])
        if save_fname is None or save_fname in ['', ()]:
            return
        if os.path.splitext(save_fname)[1] == '':
            save_fname += '.validation.txt'
        self.log_handler.save_to_file(save_fname)

    def perform_basic_validation(self):
        if self.variables.image_reader is None:
            return

        the_reader = self.variables.image_reader.base_reader
        if isinstance(the_reader, SICDReader):
            msg_id = 'SICD structure for file {}'.format(the_reader.file_name)
            self.logger.info('Starting validation of {}'.format(msg_id))
            # noinspection PyUnresolvedReferences
            result = check_file(the_reader.nitf_details)
            if result:
                self.logger.info('***{} appears to be valid***'.format(msg_id))
            self.logger.info('Completed validation for {}\n'.format(msg_id))
        else:
            # noinspection PyUnresolvedReferences
            the_sicds = the_reader.get_sicds_as_tuple()
            for the_index, the_sicd in enumerate(the_sicds):
                msg_id = 'SICD structure at index {}'.format(the_index) if len(the_sicds) > 1 else 'SICD structure'
                self.logger.info('Starting validation of {}'.format(msg_id))
                result = the_sicd.is_valid(recursive=True, stack=False)  # this implicitly logs things of note
                if result:
                    self.logger.info('***{} appears to be valid***'.format(msg_id))
                self.logger.info('Completed validation for {}\n'.format(msg_id))

    def update_reader(self, the_reader, update_browse=None):
        """
        Update the reader.

        Parameters
        ----------
        the_reader : None|str|SICDTypeReader|SICDTypeCanvasImageReader
        update_browse : None|str
        """

        if the_reader is None:
            return

        if update_browse is not None:
            self.variables.browse_directory = update_browse
        elif isinstance(the_reader, str):
            self.variables.browse_directory = os.path.split(the_reader)[0]

        if isinstance(the_reader, str):
            the_reader = SICDTypeCanvasImageReader(the_reader)

        if isinstance(the_reader, SICDTypeReader):
            the_reader = SICDTypeCanvasImageReader(the_reader)

        if not isinstance(the_reader, SICDTypeCanvasImageReader):
            raise TypeError('Got unexpected input for the reader')

        # update the reader
        self.variables.image_reader = the_reader
        self.set_title()
        # refresh appropriate GUI elements
        self.my_populate_metaicon()
        self.my_populate_metaviewer()
        self.log_handler.clear()
        # perform the initial validation
        self.logger.info(
            'Preparing validation for file {}\n'.format(
                os.path.abspath(self.variables.image_reader.file_name)))
        self.perform_basic_validation()

    def _disconnect_logging(self):
        if self.log_handler is None:
            return

        logger = logging.getLogger('validation')
        logger.removeHandler(self.log_handler)
        self.log_handler = None

    def callback_select_file(self):
        fname = askopenfilename(initialdir=self.variables.browse_directory, filetypes=common_use_collection)
        if fname is None or fname in ['', ()]:
            return

        try:
            the_reader = SICDTypeCanvasImageReader(fname)
        except SarpyIOError:
            showinfo('Opener not found',
                     message='File {} was not successfully opened as a SICD type.'.format(fname))
            return
        self.update_reader(the_reader, update_browse=os.path.split(fname)[0])

    def callback_select_directory(self):
        dirname = askdirectory(initialdir=self.variables.browse_directory, mustexist=True)
        if dirname is None or dirname in [(), '']:
            return

        try:
            the_reader = SICDTypeCanvasImageReader(dirname)
        except SarpyIOError:
            showinfo('Opener not found',
                     message='Directory {} was not successfully opened as a SICD type.'.format(dirname))
            return
        self.update_reader(the_reader, update_browse=os.path.split(dirname)[0])

    def callback_local_fs(self):
        """
        Enable the local frequency support analysis
        """

        if not self._verify_reader():
            return

        # create a complex image reader - don't pass the same one around, so no hidden state
        reader = SICDTypeCanvasImageReader(self.variables.image_reader.base_reader)
        # open the frequency support tool based on this reader
        root = tkinter.Toplevel(self.master)  # create a new toplevel with its own mainloop, so it's blocking
        tool = LocalFrequencySupportTool(root, reader=reader)
        root.grab_set()
        root.wait_window()

        self._get_and_log_feedback('Local Frequency Support (DeltaKCOAPoly)')

    def callback_full_fs(self):
        """
        Enable the full image frequency analysis
        """

        if not self._verify_reader():
            return

        # create a complex image reader - don't pass the same one around, so no hidden state
        reader = SICDTypeCanvasImageReader(self.variables.image_reader.base_reader)
        # open the frequency support tool based on this reader
        root = tkinter.Toplevel(self.master)  # create a new toplevel with its own mainloop, so it's blocking
        tool = FullFrequencySupportTool(root, reader=reader)
        root.grab_set()
        root.wait_window()

        self._get_and_log_feedback('Full Image Frequency Support')

    def callback_sign(self):
        """
        Enable Fourier sign analysis
        """

        if not self._verify_reader():
            return

        # create a complex image reader - don't pass the same one around, so no hidden state
        reader = SICDTypeCanvasImageReader(self.variables.image_reader.base_reader)
        # open the aperture tool based on this reader
        root = tkinter.Toplevel(self.master)  # create a new toplevel with its own mainloop, so it's blocking
        tool = RegionSelection(root, reader=reader)
        root.grab_set()
        root.wait_window()

        self._get_and_log_feedback('Fourier Sign')

    def callback_noise(self):
        """
        Enable noise analysis
        """

        if not self._verify_reader():
            return

        # create a complex image reader - don't pass the same one around, so no hidden state
        reader = SICDTypeCanvasImageReader(self.variables.image_reader.base_reader)
        # open the rcs tool based on this reader
        root = tkinter.Toplevel()  # create a new toplevel with its own mainloop, so it's blocking
        tool = RCSTool(root, reader=reader)
        root.grab_set()
        root.wait_window()

        self._get_and_log_feedback('Noise Value')

    def callback_geolocation(self):
        """
        Enable basic geolocation comparison
        """

        if not self._verify_reader():
            return

        # find a place to save the overlay, then produce it
        initialdir, fstem = os.path.split(os.path.abspath(self.variables.image_reader.file_name))
        dirname = askdirectory(initialdir=initialdir, mustexist=True)
        if dirname is None or dirname in [(), '']:
            return

        fstem_part = os.path.splitext(fstem)[0]
        kmz_file_stem = 'View-{}'.format(fstem_part)
        showinfo('KMZ creation',
                 message='This may be somewhat time consuming.\n'
                         'KMZ file(s) being created in directory {}\n'
                         'The created filename will begin with {}\n'
                         'Once the file(s) are created, review and provide feedback.'.format(dirname, kmz_file_stem))
        create_kmz_view(
            self.variables.image_reader.base_reader, dirname,
            inc_scp=True, inc_collection_wedge=True,
            file_stem=kmz_file_stem, pixel_limit=3072)
        showinfo('KMZ creation complete',
                 message='KMZ file(s) created in directory {}\n'
                         'The created filename(s) begin with {}\n'
                         'Review and provide feedback.'.format(dirname, kmz_file_stem))

        self._get_and_log_feedback('Geolocation')

    def my_populate_metaicon(self):
        """
        Populate the metaicon.
        """

        self.populate_metaicon(self.variables.image_reader)

    def my_populate_metaviewer(self):
        """
        Populate the metaviewer.
        """

        self.populate_metaviewer(self.variables.image_reader)

    def destroy(self):
        self._disconnect_logging()
        # noinspection PyBroadException
        try:
            super(ValidationTool, self).destroy()
        except Exception:
            pass


def main(reader=None):
    """
    Main method for initializing the tool

    Parameters
    ----------
    reader : None|str|SICDTypeReader|SICDTypeCanvasImageReader
    """

    root = tkinter.Tk()

    the_style = ttk.Style()
    the_style.theme_use('classic')

    app = ValidationTool(root, reader=reader)
    root.mainloop()


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser(
        description="Open the validation tool with optional input file.",
        formatter_class=argparse.RawTextHelpFormatter)

    parser.add_argument(
        '-i', '--input', metavar='input', default=None, help='The path to the optional image file for opening.')
    args = parser.parse_args()

    main(reader=args.input)



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages
from codecs import open

import os


# Get the long description from the README file
here = os.path.abspath(os.path.dirname(__file__))
# Get the long description from the README file
with open(os.path.join(here, 'README.md'), 'r') as f:
    long_description = f.read()


# Get the relevant setup parameters from the package
parameters = {}
with open(os.path.join(here, 'sarpy_apps', '__about__.py'), 'r') as f:
    exec(f.read(), parameters)


setup(name=parameters['__title__'],
      version=parameters['__version__'],
      description=parameters['__summary__'],
      long_description=long_description,
      long_description_content_type='text/markdown',
      packages=find_packages(exclude=('*tests*', '*examples*')),
      url=parameters['__url__'],
      author=parameters['__author__'],
      author_email=parameters['__email__'],  # The primary POC
      install_requires=[
          'numpy', 'matplotlib', 'Pillow', 'sarpy>=1.2.36', 'tk_builder>=1.1.9'],
      zip_safe=True,
      test_suite="tests",
      tests_require=[],
      classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9'
      ],
      platforms=['any'],
      license='MIT'
      )



"""
A viewer for the meta-data associated with the file readers.
"""

__classification__ = "UNCLASSIFIED"
__author__ = ("Jason Casey", "Thomas McCullough")

import tkinter

from tk_builder.widgets import basic_widgets

from sarpy.io.general.base import AbstractReader
from sarpy.io.general.nitf import NITFDetails, NITFReader
from sarpy.io.complex.base import SICDTypeReader
from sarpy.io.phase_history.base import CPHDTypeReader
from sarpy.io.received.base import CRSDTypeReader
from sarpy.io.product.base import SIDDTypeReader


def _primitive_list(the_list):
    primitive = True
    for entry in the_list:
        primitive &= isinstance(entry, (float, int, str, list))
    return primitive


class Metaviewer(basic_widgets.Treeview):
    """
    For viewing a rendering of a json compatible object.
    """

    def __init__(self, master):
        """

        Parameters
        ----------
        master : tkinter.Tk|tkinter.Toplevel
            The GUI element which is the parent
        """
        basic_widgets.Treeview.__init__(self, master)
        self.master.geometry("800x600")
        self.pack(expand=tkinter.YES, fill=tkinter.BOTH)

    def hide_on_close(self):
        """
        Sets the condition so that the close button does not destroy the tool.
        """

        self.master.protocol("WM_DELETE_WINDOW", self.close_window)

    def empty_entries(self):
        """
        Empty all entries - for the purpose of reinitializing.

        Returns
        -------
        None
        """

        self.delete(*self.get_children())

    def close_window(self):
        self.master.withdraw()

    def add_node(self, the_parent, the_key, the_value):
        """
        For the given key and value, this creates the node for the given value,
        and recursively adds children, as appropriate.

        Parameters
        ----------
        the_parent : str
            The parent key for this entry.
        the_key : str
            The key for this entry - should be unique amongst children of this parent.
        the_value : dict|list|str|int|float
            The value for this entry.

        Returns
        -------
        None
        """

        real_key = '{}_{}'.format(the_parent, the_key)
        if isinstance(the_value, list):
            if _primitive_list(the_value):
                self.insert(the_parent, "end", real_key, text="{}: {}".format(the_key, the_value))
            else:
                # add a parent node for this list
                self.insert(the_parent, "end", real_key, text=the_key)
                for i, value in enumerate(the_value):
                    # add a node for each list element
                    element_key = '{}[{}]'.format(the_key, i)
                    self.add_node(real_key, element_key, value)
        elif isinstance(the_value, dict):
            self.insert(the_parent, "end", real_key, text=the_key)
            for key in the_value:
                val = the_value[key]
                self.add_node(real_key, key, val)
        elif isinstance(the_value, float):
            self.insert(the_parent, "end", real_key, text="{0:s}: {1:0.16G}".format(the_key, the_value))
        else:
            self.insert(the_parent, "end", real_key, text="{}: {}".format(the_key, the_value))

    def populate_from_reader(self, reader):
        """
        Populate the entries from a reader implementation.

        Parameters
        ----------
        reader : AbstractReader
        """

        # empty any present entries
        self.empty_entries()

        if isinstance(reader, SICDTypeReader):
            sicds = reader.get_sicds_as_tuple()
            if sicds is None:
                return
            elif len(sicds) == 1:
                self.add_node("", "SICD", sicds[0].to_dict())
            else:
                for i, entry in enumerate(sicds):
                    self.add_node("", "SICD_{}".format(i), entry.to_dict())
        elif isinstance(reader, SIDDTypeReader):
            sidds = reader.get_sidds_as_tuple()
            if sidds is None:
                pass
            elif len(sidds) == 1:
                self.add_node("", "SIDD", sidds[0].to_dict())
            else:
                for i, entry in enumerate(sidds):
                    self.add_node("", "SIDD_{}".format(i), entry.to_dict())

            sicds = reader.sicd_meta
            if sicds is not None:
                for i, entry in enumerate(sicds):
                    self.add_node("", "SICD_{}".format(i), entry.to_dict())
        elif isinstance(reader, CPHDTypeReader):
            cphd = reader.cphd_meta
            if cphd is None:
                pass
            else:
                self.add_node("", "CPHD", cphd.to_dict())
        elif isinstance(reader, CRSDTypeReader):
            crsd = reader.crsd_meta
            if crsd is None:
                pass
            else:
                self.add_node("", "CRSD", crsd.to_dict())

        if isinstance(reader, NITFReader):
            nitf_details = reader.nitf_details  # type: NITFDetails
            self.add_node("", "NITF", nitf_details.get_headers_json())
